import unittest
from expan.core.correction import *


class CorrectionTestCase(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_benjamini_hochberg(self):
        false_discovery_rate = 0.05

        original_p_values = [0.02, 0.03, 0.002, 0.001]
        corrected_alpha = benjamini_hochberg(false_discovery_rate, original_p_values)
        self.assertAlmostEqual(corrected_alpha, 0.05)

        original_p_values = [0.01, 0.2, 0.1, 0.02]
        corrected_alpha = benjamini_hochberg(false_discovery_rate, original_p_values)
        self.assertAlmostEqual(corrected_alpha, 0.025)

    def test_benjamini_hochberg_one_p_value(self):
        false_discovery_rate = 0.05
        original_p_values = [0.1]

        corrected_alpha = benjamini_hochberg(false_discovery_rate, original_p_values)
        self.assertAlmostEqual(corrected_alpha, 0.05)

    def test_bonferroni(self):
        false_discovery_rate = 0.05
        original_p_values = [0.1, 0.03, 0.02, 0.03]

        corrected_alpha = bonferroni(false_discovery_rate, original_p_values)
        self.assertAlmostEqual(corrected_alpha, 0.0125)

    def test_bonferroni_one_p_value(self):
        false_discovery_rate = 0.05
        original_p_values = [0.1]

        corrected_alpha = bonferroni(false_discovery_rate, original_p_values)
        self.assertAlmostEqual(corrected_alpha, 0.05)



import os
import logging
import pickle
import sys
import tempfile
from os.path import dirname, join, realpath

import pandas as pd
import numpy as np
from pystan import StanModel
from scipy.stats import gaussian_kde, norm, cauchy

import expan.core.statistics as statx
from expan.core.util import drop_nan
from expan.core.results import BaseTestStatistics, SampleStatistics, EarlyStoppingTestStatistics

__location__ = realpath(join(os.getcwd(), dirname(__file__)))
logger = logging.getLogger(__name__)

cache_sampling_results = False
sampling_results = {}  # memorized sampling results


def obrien_fleming(information_fraction, alpha=0.05):
    """ Calculate an approximation of the O'Brien-Fleming alpha spending function.

    :param information_fraction: share of the information  amount at the point of evaluation, 
                                 e.g. the share of the maximum sample size
    :type  information_fraction: float
    :param alpha: type-I error rate
    :type  alpha: float

    :return: redistributed alpha value at the time point with the given information fraction
    :rtype:  float
    """
    return (1 - norm.cdf(norm.ppf(1 - alpha / 2) / np.sqrt(information_fraction))) * 2


def make_group_sequential(spending_function='obrien_fleming', estimated_sample_size=None, alpha=0.05, cap=8):
    """ A closure to the group_sequential function. """
    return lambda x, y: group_sequential(x, y, spending_function, estimated_sample_size, alpha, cap)


def group_sequential(x, y, spending_function='obrien_fleming', estimated_sample_size=None, alpha=0.05, cap=8):
    """ Group sequential method to determine whether to stop early.

    :param x: sample of a treatment group
    :type  x: pd.Series or array-like
    :param y: sample of a control group
    :type  y: pd.Series or array-like
    :param spending_function: name of the alpha spending function, currently supports only 'obrien_fleming'.
    :type  spending_function: str
    :param estimated_sample_size: sample size to be achieved towards the end of experiment
    :type  estimated_sample_size: int
    :param alpha: type-I error rate
    :type  alpha: float
    :param cap: upper bound of the adapted z-score
    :type  cap: int

    :return: results of type EarlyStoppingTestStatistics
    :rtype:  EarlyStoppingTestStatistics
    """
    # Checking if data was provided and it has correct format
    if x is None or y is None:
        raise ValueError('Please provide two non-empty samples.')
    if not isinstance(x, pd.Series) and not isinstance(x, np.ndarray) and not isinstance(x, list):
        raise TypeError('Please provide samples of type Series or list.')
    if type(x) != type(y):
        raise TypeError('Please provide samples of the same type.')

    logger.info("Started running group sequential early stopping; spending function is {}, size of treatment is {} "
                "and size of control is {}".format(spending_function, len(x), len(y)))

    # Coercing missing values to right format
    _x = np.array(x, dtype=float)
    _y = np.array(y, dtype=float)

    n_x = statx.sample_size(_x)
    n_y = statx.sample_size(_y)

    if not estimated_sample_size:
        information_fraction = 1.0
    else:
        information_fraction = min(1.0, (n_x + n_y) / estimated_sample_size)

    # alpha spending function
    if spending_function in ('obrien_fleming'):
        func = eval(spending_function)
    else:
        raise NotImplementedError
    alpha_new = func(information_fraction, alpha=alpha)

    # calculate the z-score bound
    bound = norm.ppf(1 - alpha_new / 2)
    # replace potential inf with an upper bound
    if bound == np.inf:
        bound = cap

    mu_x = np.nanmean(_x)
    mu_y = np.nanmean(_y)
    sigma_x = np.nanstd(_x)
    sigma_y = np.nanstd(_y)
    z = (mu_x - mu_y) / np.sqrt(sigma_x ** 2 / n_x + sigma_y ** 2 / n_y)

    if z > bound or z < -bound:
        stop = True
    else:
        stop = False

    interval = statx.normal_difference(mu_x, sigma_x, n_x, mu_y, sigma_y, n_y,
                                       [alpha_new * 100 / 2, 100 - alpha_new * 100 / 2])

    treatment_statistics = SampleStatistics(int(n_x), float(np.nanmean(_x)), float(np.nanvar(_x)))
    control_statistics   = SampleStatistics(int(n_y), float(np.nanmean(_y)), float(np.nanvar(_y)))
    variant_statistics   = BaseTestStatistics(control_statistics, treatment_statistics)
    p_value              = statx.compute_p_value_from_samples(_x, _y)
    statistical_power    = statx.compute_statistical_power_from_samples(_x, _y, alpha)

    logger.info("Finished running group sequential early stopping; spending function is {}, size of treatment is {} "
                "and size of control is {}".format(spending_function, len(x), len(y)))
    return EarlyStoppingTestStatistics(variant_statistics.control_statistics,
                                       variant_statistics.treatment_statistics,
                                       float(mu_x - mu_y), interval, p_value, statistical_power, stop)


def HDI_from_MCMC(posterior_samples, credible_mass=0.95):
    """ Computes highest density interval from a sample of representative values, 
    estimated as the shortest credible interval.
    Takes Arguments posterior_samples (samples from posterior) and credible mass (normally .95).
    http://stackoverflow.com/questions/22284502/highest-posterior-density-region-and-central-credible-region
    
    :param posterior_samples: sample of data points from posterior distribution of some parameter
    :type  posterior_samples: array-like
    :param credible_mass: the range of credible interval. 0.95 means 95% represents credible interval.
    :type  credible_mass: float

    :return: corresponding lower and upper bound for the credible interval
    :rtype:  tuple[float]
    """

    sorted_points = sorted(posterior_samples)
    ciIdxInc = np.ceil(credible_mass * len(sorted_points)).astype('int')
    nCIs = len(sorted_points) - ciIdxInc
    ciWidth = [0] * nCIs
    for i in range(0, nCIs):
        ciWidth[i] = sorted_points[i + ciIdxInc] - sorted_points[i]

    HDImin = sorted_points[ciWidth.index(min(ciWidth))]
    HDImax = sorted_points[ciWidth.index(min(ciWidth)) + ciIdxInc]
    return (HDImin, HDImax)


def _bayes_sampling(x, y, distribution='normal', num_iters=25000, inference="sampling"):
    """ Helper function for bayesian sampling.

    :param x: sample of a treatment group
    :type  x: pd.Series or list (array-like)
    :param y: sample of a control group
    :type  y: pd.Series or list (array-like)
    :param distribution: name of the KPI distribution model, which assumes a Stan model file with the same name exists
    :type  distribution: str
    :param num_iters: number of iterations of sampling
    :type  num_iters: int
    :param inference: 'sampling' for MCMC sampling method or 'variational' for variational inference
    :type  inference: str

    :return: the posterior samples, sample size of x, sample size of y, absolute mean of x, absolute mean of y
    :rtype:  tuple[array-like, array-like, array-like, float, float]
    """
    # Checking if data was provided and it has correct format
    if x is None or y is None:
        raise ValueError('Please provide two non-empty samples.')
    if not isinstance(x, pd.Series) and not isinstance(x, np.ndarray) and not isinstance(x, list):
        raise TypeError('Please provide samples of type Series or list.')
    if type(x) != type(y):
        raise TypeError('Please provide samples of the same type.')

    logger.info("Started running bayesian inference with {} procedure, treatment group of size {}, "
                "control group of size {}, {} distribution.".format(inference, len(x), len(y), distribution, inference))

    # Coercing missing values to right format
    _x = np.array(x, dtype=float)
    _y = np.array(y, dtype=float)
    _x = drop_nan(_x)
    _y = drop_nan(_y)

    key = (str(_x), str(_y), num_iters, inference)

    if cache_sampling_results and key in sampling_results:
        return sampling_results[key]

    mu_x = np.nanmean(_x)
    mu_y = np.nanmean(_y)
    n_x = statx.sample_size(_x)
    n_y = statx.sample_size(_y)

    if distribution == 'normal':
        fit_data = {'Nc': n_y,
                    'Nt': n_x,
                    'x': _x,
                    'y': _y}
    elif distribution == 'poisson':
        fit_data = {'Nc': n_y,
                    'Nt': n_x,
                    'x': _x.astype(int),
                    'y': _y.astype(int)}
    else:
        raise NotImplementedError

    model_file = __location__ + '/../models/' + distribution + '_kpi.stan'

    sm = get_or_compile_stan_model(model_file, distribution)

    if inference == "sampling":
        fit = sm.sampling(data=fit_data, iter=num_iters, chains=4, n_jobs=1, seed=1,
                          control={'stepsize': 0.01, 'adapt_delta': 0.99})
        traces = fit.extract()

    elif inference == "variational":
        results_dict = sm.vb(data=fit_data, iter=10000)
        traces = {}
        for i in range(len(results_dict['sampler_param_names'])):
            para_name = results_dict['sampler_param_names'][i]
            para_values = np.array(results_dict['sampler_params'][i])
            traces[para_name] = para_values

    if cache_sampling_results:
        sampling_results[key] = (traces, n_x, n_y, mu_x, mu_y)

    logger.info("Finished running bayesian inference with {} procedure, treatment group of size {}, "
                "control group of size {}, {} distribution.".format(inference, len(x), len(y), distribution))
    return traces, n_x, n_y, mu_x, mu_y


def make_bayes_factor(distribution='normal', num_iters=25000, inference='sampling'):
    """ Closure method for the bayes_factor"""
    def f(x, y):
        return bayes_factor(x, y, distribution, num_iters, inference)
    return f


def bayes_factor(x, y, distribution='normal', num_iters=25000, inference='sampling'):
    """ Bayes factor computation.

    :param x: sample of a treatment group
    :type  x: pd.Series or list (array-like)
    :param y: sample of a control group
    :type  y: pd.Series or list (array-like)
    :param distribution: name of the KPI distribution model, which assumes a Stan model file with the same name exists
    :type  distribution: str
    :param num_iters: number of iterations of bayes sampling
    :type  num_iters: int
    :param inference: sampling or variational inference method for approximation the posterior
    :type  inference: str

    :return: results of type EarlyStoppingTestStatistics (without p-value and stat. power)
    :rtype:  EarlyStoppingTestStatistics
    """

    logger.info("Started running bayes factor with {} procedure, treatment group of size {}, "
                "control group of size {}, {} distribution.".format(len(x), len(y), distribution, inference))

    traces, n_x, n_y, mu_x, mu_y = _bayes_sampling(x, y, distribution=distribution, num_iters=num_iters,
                                                   inference=inference)
    trace_normalized_effect_size = get_trace_normalized_effect_size(distribution, traces)
    trace_absolute_effect_size = traces['delta']

    kde = gaussian_kde(trace_normalized_effect_size)
    prior = cauchy.pdf(0, loc=0, scale=1)
    # BF_01
    bf = kde.evaluate(0)[0] / prior
    stop = bf > 3 or bf < 1 / 3.

    credible_mass = 0.95
    left_out      = 1.0 - credible_mass
    p1            = round(left_out/2.0, 5)
    p2            = round(1.0 - left_out/2.0, 5)
    credible_interval = HDI_from_MCMC(trace_absolute_effect_size, credible_mass)

    treatment_statistics = SampleStatistics(int(n_x), float(mu_x), float(np.nanvar(x)))
    control_statistics   = SampleStatistics(int(n_y), float(mu_y), float(np.nanvar(y)))
    variant_statistics   = BaseTestStatistics(control_statistics, treatment_statistics)

    logger.info("Finished running bayes factor with {} procedure, treatment group of size {}, "
                "control group of size {}, {} distribution.".format(len(x), len(y), distribution, inference))

    return EarlyStoppingTestStatistics(variant_statistics.control_statistics,
                                       variant_statistics.treatment_statistics,
                                       float(mu_x - mu_y),
                                       dict([(p*100, v) for p, v in zip([p1, p2], credible_interval)]),
                                       None, None, stop)


def make_bayes_precision(distribution='normal', posterior_width=0.08, num_iters=25000, inference='sampling'):
    """ Closure method for the bayes_precision"""
    def f(x, y):
        return bayes_precision(x, y, distribution, posterior_width, num_iters, inference)
    return f


def bayes_precision(x, y, distribution='normal', posterior_width=0.08, num_iters=25000, inference='sampling'):
    """ Bayes precision computation.

    :param x: sample of a treatment group
    :type  x: pd.Series or list (array-like)
    :param y: sample of a control group
    :type  y: pd.Series or list (array-like)
    :param distribution: name of the KPI distribution model, which assumes a Stan model file with the same name exists
    :type  distribution: str
    :param posterior_width: the stopping criterion, threshold of the posterior  width
    :type  posterior_width: float
    :param num_iters: number of iterations of bayes sampling
    :type  num_iters: int
    :param inference: sampling or variational inference method for approximation the posterior
    :type  inference: str

    :return: results of type EarlyStoppingTestStatistics (without p-value and stat. power)
    :rtype:  EarlyStoppingTestStatistics
    """

    logger.info("Started running bayes precision with {} procedure, treatment group of size {}, "
                "control group of size {}, {} distribution.".format(len(x), len(y), distribution, inference))

    traces, n_x, n_y, mu_x, mu_y = _bayes_sampling(x, y, distribution=distribution,
                                                   num_iters=num_iters, inference=inference)
    trace_normalized_effect_size = get_trace_normalized_effect_size(distribution, traces)
    trace_absolute_effect_size = traces['delta']

    credible_mass = 0.95
    left_out      = 1.0 - credible_mass
    p1            = round(left_out/2.0, 5)
    p2            = round(1.0 - left_out/2.0, 5)

    credible_interval_delta            = HDI_from_MCMC(trace_absolute_effect_size, credible_mass)
    credible_interval_delta_normalized = HDI_from_MCMC(trace_normalized_effect_size, credible_mass)

    stop = credible_interval_delta_normalized[1] - credible_interval_delta_normalized[0] < posterior_width

    treatment_statistics = SampleStatistics(int(n_x), float(mu_x), float(np.nanvar(x)))
    control_statistics   = SampleStatistics(int(n_y), float(mu_y), float(np.nanvar(y)))
    variant_statistics   = BaseTestStatistics(control_statistics, treatment_statistics)

    logger.info("Finished running bayes precision with {} procedure, treatment group of size {}, "
                "control group of size {}, {} distribution.".format(len(x), len(y), distribution, inference))

    return EarlyStoppingTestStatistics(variant_statistics.control_statistics,
                                       variant_statistics.treatment_statistics,
                                       float(mu_x - mu_y),
                                       dict([(p * 100, v) for p, v in zip([p1, p2], credible_interval_delta)]),
                                       None, None, stop)


def get_trace_normalized_effect_size(distribution, traces):
    """ Obtaining a Stan model statistics for 'normal' or 'poisson' distribution

    :param distribution: name of the KPI distribution model, which assumes a Stan model file with the same name exists
    :type  distribution: str
    :param traces: sampling statistics
    :type  traces: dict

    :return: sample of data points from posterior distribution of some parameter
    :rtype:  array-like
    """
    # check of traces type, if traces is not dict "traces['alpha']" will return KeyError
    if not isinstance(traces, dict):
        raise TypeError("Traces statistics is not a dictionary and does not contain alpha or delta statistics.")

    if distribution == 'normal':
        return traces['alpha']
    elif distribution == 'poisson':
        variance = np.nanmean(np.array(traces['delta']))
        return traces['delta'] / np.sqrt(np.absolute(variance))
    else:
        raise ValueError("Model " + distribution + " is not implemented.")


def get_or_compile_stan_model(model_file, distribution):
    """ Creates Stan model. Compiles a Stan model and saves it to .pkl file to the folder selected by tempfile module if
    file doesn't exist yet and load precompiled model if there is a model file in temporary dir.
    
    Note: compiled_model_file is the hardcoded file path which may cause some issues in future.
    There are 2 alternative implementations for Stan models handling:

    1. Using global variables
    2. Pre-compiling stan models and adding them as a part of ExpAn project

    Using temporary files with tempfile module is not currently possible,
    since it generates a unique file name which is difficult to track.
    However, compiled modules are saved in temporary directory using tempfile module 
    which vary based on the current platform and settings. Cleaning up a temp dir is done on boot.

    :param model_file: model file location
    :type  model_file: str
    :param distribution: name of the KPI distribution model, which assumes a Stan model file with the same name exists
    :type  distribution: str

    :return: compiled Stan model for the selected distribution or normal distribution as a default option
    :rtype:  Class representing a compiled Stan model
    """

    logger.info("Started loading and compiling Stan model for {} distribution".format(distribution))

    if distribution is not 'normal' and distribution is not 'poisson':
        raise ValueError("Model " + distribution + " is not implemented.")

    python_version = '{0[0]}.{0[1]}'.format(sys.version_info)
    compiled_model_file = tempfile.gettempdir() + '/expan_early_stop_compiled_stan_model_' \
                                                + distribution + '_' + python_version + '.pkl'

    if os.path.isfile(compiled_model_file):
        sm = pickle.load(open(compiled_model_file, 'rb'))
    else:
        sm = StanModel(file=model_file)
        with open(compiled_model_file, 'wb') as f:
            pickle.dump(sm, f)
    return sm



import unittest

import numpy as np

import expan.core.early_stopping as es
from expan.core.util import find_value_by_key_with_condition


class EarlyStoppingTestCase(unittest.TestCase):
    """ Defines the setUp() and tearDown() functions for the early-stopping test cases."""

    def setUp(self):
        """ Load the needed datasets for all EarlyStoppingTestCases and set the random
        seed so that randomized algorithms show deterministic behaviour.
        """
        np.random.seed(0)

        self.rand_s1 = np.random.normal (loc=0.0, size=1000)
        self.rand_s2 = np.random.normal (loc=0.1, size=1000)
        self.rand_s3 = np.random.poisson(lam=1.0, size=1000)
        self.rand_s4 = np.random.poisson(lam=3.0, size=1000)
        self.rand_s5 = np.random.normal (loc=0.0, size=1000)
        self.rand_s6 = np.random.normal (loc=0.1, size=1000)
        self.rand_s5[0] = np.nan
        self.rand_s6[0] = np.nan
        self.rand_s6[1] = np.nan

    def tearDown(self):
        """Clean up after the test"""
        pass


class GroupSequentialTestCases(EarlyStoppingTestCase):
    """ Test cases for the group sequential functions in core.early_stopping."""

    def test_obrien_fleming(self):
        """ Check the O'Brien-Fleming spending function."""
        # Check array as input
        res_1 = es.obrien_fleming(np.linspace(0, 1, 5 + 1)[1:])
        expected_res = [1.17264468e-05, 1.94191300e-03, 1.13964185e-02, 2.84296308e-02, 5.00000000e-02]
        np.testing.assert_almost_equal(res_1, expected_res)

        # Check float as input
        res_2 = es.obrien_fleming(0.5)
        self.assertAlmostEqual(res_2, 0.005574596680784305)

        # Check int as input
        res_3 = es.obrien_fleming(1)
        self.assertAlmostEqual(res_3, 0.05)

    def test_group_sequential(self):
        """ Check the group sequential function."""
        res = es.group_sequential(self.rand_s1, self.rand_s2)

        self.assertEqual(res.treatment_statistics.sample_size,     1000)
        self.assertEqual(res.control_statistics.sample_size,       1000)
        self.assertAlmostEqual(res.treatment_statistics.mean,     -0.045256707490195384)
        self.assertAlmostEqual(res.control_statistics.mean,        0.11361694031616358)
        self.assertAlmostEqual(res.treatment_statistics.variance,  0.9742344563121542)
        self.assertAlmostEqual(res.control_statistics.variance,    0.9373337542827797)

        self.assertAlmostEqual(res.delta, -0.15887364780635896)
        value025 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 2.5, 'value')
        value975 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 97.5, 'value')
        np.testing.assert_almost_equal(value025, -0.24461812530841959, decimal=5)
        np.testing.assert_almost_equal(value975, -0.07312917030429833, decimal=5)

        self.assertAlmostEqual(res.p,                 0.0002863669955157941)
        self.assertAlmostEqual(res.statistical_power, 0.9529152504960496)
        self.assertEqual(res.stop,                    True)

    def test_group_sequential_actual_size_larger_than_estimated(self):
        """ Check the group sequential function with wrong input,
        such that the actual data size is already larger than estimated sample size.
        """
        res = es.group_sequential(self.rand_s1, self.rand_s2, estimated_sample_size=100)

        value025 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 2.5, 'value')
        value975 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 97.5, 'value')
        np.testing.assert_almost_equal (value025, -0.24461812530841959, decimal=5)
        np.testing.assert_almost_equal (value975, -0.07312917030429833, decimal=5)


class BayesFactorTestCases(EarlyStoppingTestCase):
    """ Test cases for the bayes_factor function in core.early_stopping."""

    # @unittest.skip("sometimes takes too much time")
    def test_bayes_factor(self):
        """ Check the Bayes factor function."""
        res = es.bayes_factor(self.rand_s1, self.rand_s2, num_iters=2000)

        self.assertEqual(res.treatment_statistics.sample_size,     1000)
        self.assertEqual(res.control_statistics.sample_size,       1000)
        self.assertAlmostEqual(res.treatment_statistics.mean,     -0.045256707490195384)
        self.assertAlmostEqual(res.control_statistics.mean,        0.11361694031616358)
        self.assertAlmostEqual(res.treatment_statistics.variance,  0.9742344563121542)
        self.assertAlmostEqual(res.control_statistics.variance,    0.9373337542827797)

        self.assertAlmostEqual(res.delta, -0.15887364780635896)
        value025 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 2.5, 'value')
        value975 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 97.5, 'value')
        np.testing.assert_almost_equal(value025, -0.24293384641452503, decimal=5)
        np.testing.assert_almost_equal(value975, -0.075064346336461404, decimal=5)
        self.assertEqual(res.p,                 None)
        self.assertEqual(res.statistical_power, None)
        self.assertEqual(res.stop,              True)

    # @unittest.skip("sometimes takes too much time")
    def test_bayes_factor_poisson(self):
        """ Check the Bayes factor function for Poisson distributions."""
        res = es.bayes_factor(self.rand_s3, self.rand_s4, distribution='poisson', num_iters=2000)

        self.assertEqual(res.treatment_statistics.sample_size,      1000)
        self.assertEqual(res.control_statistics.sample_size,        1000)
        self.assertAlmostEqual(res.treatment_statistics.mean,       0.96599999999999997)
        self.assertAlmostEqual(res.control_statistics.mean,         2.9249999999999998)
        self.assertAlmostEqual(res.treatment_statistics.variance,   0.868844)
        self.assertAlmostEqual(res.control_statistics.variance,     2.901375)

        self.assertAlmostEqual(res.delta, -1.9589999999999999)
        value025 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 2.5, 'value')
        value975 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 97.5, 'value')
        np.testing.assert_almost_equal(value025, -2.0713281392132465, decimal=5)
        np.testing.assert_almost_equal(value975, -1.8279692168150592, decimal=5)
        self.assertEqual(res.p,                 None)
        self.assertEqual(res.statistical_power, None)
        self.assertEqual(res.stop,              True)

    # @unittest.skip("sometimes takes too much time")
    def test_bayes_factor_with_nan_input(self):
        """ Check the Bayes factor function with input that contains nan values."""
        res = es.bayes_factor(self.rand_s5, self.rand_s6, num_iters=2000)
        self.assertEqual(res.stop, True)

    def test_variational_inference(self):
        """ Check bayesian sampling using variational bayes."""
        traces, n_x, n_y, mu_x, mu_y = es._bayes_sampling(self.rand_s1, self.rand_s2,
                                                          num_iters=2000, inference="variational")
        self.assertEqual(len(traces), 4)
        self.assertEqual(len(traces['delta']), 1001)
        self.assertEqual(n_x, 1000)
        self.assertEqual(n_y, 1000)


class BayesPrecisionTestCases(EarlyStoppingTestCase):
    """ Test cases for the bayes_precision function in core.early_stopping."""

    # @unittest.skip("sometimes takes too much time")
    def test_bayes_precision(self):
        """ Check the bayes_precision function."""
        res = es.bayes_precision(self.rand_s1, self.rand_s2, num_iters=2000)

        self.assertEqual(res.treatment_statistics.sample_size, 1000)
        self.assertEqual(res.control_statistics.sample_size, 1000)
        self.assertAlmostEqual(res.treatment_statistics.mean, -0.045256707490195384)
        self.assertAlmostEqual(res.control_statistics.mean, 0.11361694031616358)
        self.assertAlmostEqual(res.treatment_statistics.variance, 0.9742344563121542)
        self.assertAlmostEqual(res.control_statistics.variance, 0.9373337542827797)

        self.assertAlmostEqual(res.delta, -0.15887364780635896)
        value025 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 2.5, 'value')
        value975 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 97.5, 'value')
        np.testing.assert_almost_equal(value025, -0.24293384641452503, decimal=5)
        np.testing.assert_almost_equal(value975, -0.07506434633646140, decimal=5)
        self.assertEqual(res.p,                 None)
        self.assertEqual(res.statistical_power, None)
        self.assertEqual(res.stop,              False)

if __name__ == '__main__':
    unittest.main()



# TODO: This module is deprecated

import sys
import unittest

import pandas as pd

from expan.core.binning import *

#---------- test util function ------------#
def toBinRepresentation(bins):
    """
    Get the bin representations of the given bins
    :param bins: a list of Bin objects.
    :return: a lista list of NumericalRepresenation or CategoricalRepresenation objects
    """
    reprs = []
    for bin in bins:
        reprs.append(bin.representation)
    return reprs


#---------- test classes ------------#
class BinningTestCase(unittest.TestCase):
    """
    Defines the setUp() and tearDown() functions for the statistics test cases.
    """
    def setUp(self):
        np.random.seed(41)
        self.maxDiff = None

    def tearDown(self):
        pass

    def assertCollectionEqual(self, source, expected):
        is_python3 = sys.version_info[0] == 3
        if is_python3:
            return self.assertCountEqual(source, expected)
        else:
            return self.assertItemsEqual(source, expected)



#---------- Numerical binning tests ------------#
class CreateNumericalBinsTestCase(BinningTestCase):
    """
    Test cases for creating numerical bins.
    """
    def test_invalid_input(self):
        with self.assertRaises(ValueError):
            create_bins(None, 10)
        with self.assertRaises(ValueError):
            create_bins(list(range(100)), 0)

    def test_create_regular(self):
        data = np.arange(1000)
        nbins = 10
        bins = create_bins(data, nbins)
        self.assertEqual(len(bins), nbins)
        bins_repr_source = toBinRepresentation(bins)
        bins_repr_expected = [
            NumericalRepresentation(0, 100, True, False),
            NumericalRepresentation(100, 200, True, False),
            NumericalRepresentation(200, 300, True, False),
            NumericalRepresentation(300, 400, True, False),
            NumericalRepresentation(400, 500, True, False),
            NumericalRepresentation(500, 600, True, False),
            NumericalRepresentation(600, 700, True, False),
            NumericalRepresentation(700, 800, True, False),
            NumericalRepresentation(800, 900, True, False),
            NumericalRepresentation(900, 999, True, True)
        ]
        self.assertCollectionEqual(bins_repr_source, bins_repr_expected)

    def test_create_nan(self):
        data = np.arange(1002.)
        data[-2] = np.nan
        data[-1] = np.nan
        nbins = 11
        bins = create_bins(data, nbins)
        self.assertEqual(len(bins), nbins)
        bins_repr_source = toBinRepresentation(bins)
        bins_repr_expected = [
            NumericalRepresentation(0, 100, True, False),
            NumericalRepresentation(100, 200, True, False),
            NumericalRepresentation(200, 300, True, False),
            NumericalRepresentation(300, 400, True, False),
            NumericalRepresentation(400, 500, True, False),
            NumericalRepresentation(500, 600, True, False),
            NumericalRepresentation(600, 700, True, False),
            NumericalRepresentation(700, 800, True, False),
            NumericalRepresentation(800, 900, True, False),
            NumericalRepresentation(900, 999, True, True),
            NumericalRepresentation(np.nan, np.nan, True, True)  # bin for nans
        ]
        self.assertCollectionEqual(bins_repr_source, bins_repr_expected)

    def test_creation_two_big_bins_noise_between(self):
        data = [0] * 10000 + list(range(300)) + [301] * 10000
        nbins = 10
        with warnings.catch_warnings(record=True) as w:
            bins = create_bins(data, nbins)
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, UserWarning))
            self.assertTrue('less bins than requested' in str(w[-1].message).lower())
        bins_repr_source = toBinRepresentation(bins)
        bins_repr_expected = [
            NumericalRepresentation(0, 0, True, True),
            NumericalRepresentation(1, 301, True, False),
            NumericalRepresentation(301, 301, True, True)
        ]
        self.assertCollectionEqual(bins_repr_source, bins_repr_expected)

    def test_creation_very_skewed_data(self):
        data = [0] * 10000 + list(range(300))
        nbins = 4
        bins = create_bins(data, nbins)
        bins_repr_source = toBinRepresentation(bins)
        bins_repr_expected = [
            NumericalRepresentation(0, 0, True, True),
            NumericalRepresentation(1, 101, True, False),
            NumericalRepresentation(101, 200, True, False),
            NumericalRepresentation(200, 299, True, True)
        ]
        self.assertCollectionEqual(bins_repr_source, bins_repr_expected)

    def test_creation_single_bin(self):
        data = [0] * 100
        nbins = 1
        bins = create_bins(data, nbins)
        bins_repr_source = toBinRepresentation(bins)
        bins_repr_expected = [NumericalRepresentation(0, 0, True, True)]
        self.assertCollectionEqual(bins_repr_source, bins_repr_expected)

    def test_creation_more_bins_than_data(self):
        data = [0] * 100 + [1] * 10
        nbins = 3
        with warnings.catch_warnings(record=True) as w:
            bins = create_bins(data, nbins)
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, UserWarning))
            self.assertTrue('unique values' in str(w[-1].message))
        bins_repr_source = toBinRepresentation(bins)
        bins_repr_expected = [
            NumericalRepresentation(0, 0, True, True),
            NumericalRepresentation(1, 1, True, True)
        ]
        self.assertCollectionEqual(bins_repr_source, bins_repr_expected)

    def test_creation_range_100__n_bins_8(self):
        data = list(range(100))
        nbins = 8
        bins = create_bins(data, nbins)
        bins_repr_source = toBinRepresentation(bins)
        bins_repr_expected = [
            NumericalRepresentation(0, 13, True, False),
            NumericalRepresentation(13, 26, True, False),
            NumericalRepresentation(26, 39, True, False),
            NumericalRepresentation(39, 51, True, False),
            NumericalRepresentation(51, 63, True, False),
            NumericalRepresentation(63, 76, True, False),
            NumericalRepresentation(76, 88, True, False),
            NumericalRepresentation(88, 99, True, True)
        ]
        self.assertCollectionEqual(bins_repr_source, bins_repr_expected)

    def test_creation_range_100__n_bins_2(self):
        data = list(range(100))
        nbins = 2
        bins = create_bins(data, nbins)
        bins_repr_source = toBinRepresentation(bins)
        bins_repr_expected = [
            NumericalRepresentation(0, 50, True, False),
            NumericalRepresentation(50, 99, True, True)
        ]
        self.assertCollectionEqual(bins_repr_source, bins_repr_expected)


class ApplyNumericalBinsTestCase(BinningTestCase):
    """
    Test cases for applying bins to numerical data.
    """
    def test_assign_regular(self):
        data = pd.DataFrame(np.tile(np.array([np.arange(1000)]).T, (1,3)), columns=list('ABC'))
        dimension = 'A'

        bin1 = Bin("numerical", 0, 10, True, False)
        data_applied_bin1 = pd.DataFrame(np.tile(np.array([np.arange(10)]).T, (1, 3)), columns=list('ABC'))
        np.testing.assert_array_equal(data_applied_bin1, bin1(data, dimension))

        bin2 = Bin("numerical", 40, 50, True, False)
        data_applied_bin2 = pd.DataFrame(np.tile(np.array([np.arange(40, 50)]).T, (1, 3)), columns=list('ABC'))
        np.testing.assert_array_equal(data_applied_bin2, bin2(data, dimension))

        bin3 = Bin("numerical", 300, 500, False, True)
        data_applied_bin3 = pd.DataFrame(np.tile(np.array([np.arange(301, 501)]).T, (1, 3)), columns=list('ABC'))
        np.testing.assert_array_equal(data_applied_bin3, bin3(data, dimension))

        bin4 = Bin("numerical", 900, 999, True, True)
        data_applied_bin4 = pd.DataFrame(np.tile(np.array([np.arange(900, 1000)]).T, (1, 3)), columns=list('ABC'))
        np.testing.assert_array_equal(data_applied_bin4, bin4(data, dimension))

    def test_assign_unseen_data(self):
        data = pd.DataFrame(np.tile(np.array([np.arange(1000)]).T, (1,3)), columns=list('ABC'))
        dimension = 'A'

        bin = Bin("numerical", 1000, 2000, False, False)
        np.testing.assert_array_equal(None, bin(data, dimension))

    def test_assign_nan(self):
        data_one_dim = np.arange(1002.)
        data_one_dim[-2] = np.nan
        data_one_dim[-1] = np.nan
        data = pd.DataFrame(np.tile(np.array([data_one_dim]).T, (1,3)), columns=list('ABC'))
        dimension = 'A'

        bin_regular = Bin("numerical", 0, 10, True, False)
        data_applied_bin_regular = pd.DataFrame(np.tile(np.array([np.arange(10)]).T, (1, 3)), columns=list('ABC'))
        np.testing.assert_array_equal(data_applied_bin_regular, bin_regular(data, dimension))

        bin_nan = Bin("numerical", np.nan, np.nan, True, True)
        data_applied_bin_nan = pd.DataFrame( np.full((2,3), np.nan), columns=list('ABC'))
        np.testing.assert_array_equal(data_applied_bin_nan, bin_nan(data, dimension))


#---------- Categorical binning tests ------------#
class CreateCategoricalBinsTestCase(BinningTestCase):
    """
    Test cases for creating categorical bins.
    """
    def test_categorical_binning_1(self):
        data = ['a'] * 10 + ['b'] * 10
        bins = create_bins(data, 2)
        bins_repr_source = toBinRepresentation(bins)
        bins_repr_expected = [
            CategoricalRepresentation(["a"]),
            CategoricalRepresentation(["b"])
        ]
        self.assertCollectionEqual(bins_repr_source, bins_repr_expected)

    def test_categorical_binning_2(self):
        data = ['a'] * 10 + ['b'] * 5 + ['c'] * 5
        bins = create_bins(data, 2)
        bins_repr_source = toBinRepresentation(bins)
        bins_repr_expected = [
            CategoricalRepresentation(["a"]),
            CategoricalRepresentation(["c", "b"])
        ]
        self.assertCollectionEqual(bins_repr_source, bins_repr_expected)

    def test_categorical_binning_too_little_data(self):
        data = ['a'] * 10 + ['b'] * 5 + ['c'] * 5
        bins = create_bins(data, 4)
        self.assertEqual(len(bins), 3)

    def test_binning_date_1(self):
        data = ['2017-05-01'] * 10 + ['2017-06-01'] * 5 + ['2017-07-01'] * 5

        bins = create_bins(data, 4)
        self.assertEqual(len(bins), 3)

    def test_binning_date_2(self):
        data = ['2017-05-01'] * 10 + ['2017-06-01'] * 5 + ['2017-07-01'] * 5

        bins = create_bins(data, 2)
        self.assertEqual(len(bins), 2)

        bins_repr_source = toBinRepresentation(bins)

        bins_repr_expected = [
            CategoricalRepresentation(["2017-05-01"]),
            CategoricalRepresentation(["2017-07-01", "2017-06-01"])
        ]
        self.assertCollectionEqual(bins_repr_source, bins_repr_expected)


class ApplyCategoricalBinsTestCase(BinningTestCase):
    """
    Test cases for applying bins to categorical data.
    """
    def test_assign_regular(self):
        data = pd.DataFrame(["a", "b", "c", "a", "b"], columns=list("a"))

        bin_a = Bin("categorical", ["a"])
        np.testing.assert_array_equal(np.array([["a", "a"]]).T, bin_a(data, "a"))

        bin_b = Bin("categorical", ["b"])
        np.testing.assert_array_equal(np.array([["b", "b"]]).T, bin_b(data, "a"))

        bin_c = Bin("categorical", ["c"])
        np.testing.assert_array_equal(np.array([["c"]]), bin_c(data, "a"))

    def test_assign_unseen(self):
        data = pd.DataFrame(["a", "b", "c"], columns=list("a"))

        bin = Bin("categorical", ["d"])
        np.testing.assert_array_equal(None, bin(data, "a"))

    def test_assign_multiple(self):
        data = pd.DataFrame(["a", "b", "c", "a", "b"], columns=list("a"))

        bin = Bin("categorical", ["a", "b"])
        np.testing.assert_array_equal(np.array([["a", "b", "a", "b"]]).T, bin(data, "a"))



import unittest
import numpy as np

from expan.core.results import *
from expan.core.statistical_test import *
from tests.tests_core.util import get_test_data_revenue_order


class StatisticalTestCase(unittest.TestCase):
    def setUp(self):
        np.random.seed(41)
        self.data = get_test_data_revenue_order()

    def tearDown(self):
        pass

    def test_serialize_single_result(self):
        kpi = KPI('revenue')
        variants = Variants('variant', 'control', 'treatment')
        control_statistics   = SampleStatistics(1000, 0.0, 1.0)
        treatment_statistics = SampleStatistics(1200, 1.0, 1.0)
        delta                = 1.0
        p                    = 0.04
        statistical_power    = 0.8
        confidence_interval  = {2.5: 0.1, 97.5: 1.1}

        statistical_test = StatisticalTest(self.data, kpi, [], variants)
        simple_stats = SimpleTestStatistics(control_statistics,
                                            treatment_statistics,
                                            delta,
                                            confidence_interval,
                                            p,
                                            statistical_power)
        statistical_test_result = StatisticalTestResult(statistical_test, simple_stats)

        js_result = statistical_test_result.toJson()  # no error/exception should be raise
        print(js_result)    # use pytest -s to check the output if needed

    def test_serialize_multi_test_suite_result(self):
        kpi = KPI('revenue')
        variants = Variants('variant', 'control', 'treatment')
        control_statistics   = SampleStatistics(1000, 0.0, 1.0)
        treatment_statistics = SampleStatistics(1200, 1.0, 1.0)
        delta                = 1.0
        p                    = 0.04
        statistical_power    = 0.8
        confidence_interval  = {2.5: 0.1, 97.5: 1.1}
        stop                 = True

        corrected_p          = 0.02
        corrected_ci         = {1.0: -0.7, 99.0: 0.7}
        corrected_stop       = False

        simple_stats = SimpleTestStatistics(control_statistics,
                                            treatment_statistics,
                                            delta,
                                            confidence_interval,
                                            p,
                                            statistical_power)

        simple_stats_corrected = SimpleTestStatistics(control_statistics,
                                                      treatment_statistics,
                                                      delta,
                                                      corrected_ci,
                                                      corrected_p,
                                                      statistical_power)

        es_stats = EarlyStoppingTestStatistics(control_statistics,
                                               treatment_statistics,
                                               delta,
                                               confidence_interval,
                                               p,
                                               statistical_power,
                                               stop)

        es_stats_corrected = EarlyStoppingTestStatistics(control_statistics,
                                                         treatment_statistics,
                                                         delta,
                                                         corrected_ci,
                                                         corrected_p,
                                                         statistical_power,
                                                         corrected_stop)

        mobile = FeatureFilter('device_type', 'mobile')
        desktop = FeatureFilter('device_type', 'desktop')
        statistical_test_mobile = StatisticalTest(self.data, kpi, [mobile], variants)
        statistical_test_desktop = StatisticalTest(self.data, kpi, [desktop], variants)

        test_result1 = StatisticalTestResult(statistical_test_mobile, CombinedTestStatistics(simple_stats, simple_stats_corrected))
        test_result2 = StatisticalTestResult(statistical_test_desktop, CombinedTestStatistics(es_stats, es_stats_corrected))
        test_results = [test_result1, test_result2]
        statistical_test_results = MultipleTestSuiteResult(test_results, CorrectionMethod.BH)

        js_result = statistical_test_results.toJson()  # no error/exception should be raise
        print(js_result)    # use pytest -s to check the output if needed



import warnings
import logging

import numpy as np
import pandas as pd
from scipy import stats
from expan.core.results import BaseTestStatistics, SampleStatistics, SimpleTestStatistics

logger = logging.getLogger(__name__)


def _delta_mean(x, y):
    """ Calculate the delta of the two groups. Implemented as function to allow being called from bootstrap. """
    return np.nanmean(x) - np.nanmean(y)


def make_delta(assume_normal=True, alpha=0.05, min_observations=20, nruns=10000, relative=False):
    """ A closure to the delta function. """
    return lambda x, y: delta(x, y, assume_normal, alpha, min_observations, nruns, relative)


def delta(x, y, assume_normal=True, alpha=0.05, min_observations=20, nruns=10000, relative=False):
    """ Calculates the difference of means between the samples in a statistical sense.
    Computation is done in form of treatment minus control, i.e. x-y.
    Note that NaNs are treated as if they do not exist in the data. 
    
    :param x: sample of the treatment group
    :type  x: pd.Series or array-like
    :param y: sample of the control group
    :type  y: pd.Series or array-like
    :param assume_normal: specifies whether normal distribution assumptions can be made
    :type  assume_normal: boolean
    :param alpha: significance level (alpha)
    :type  alpha: float
    :param min_observations: minimum number of observations needed
    :type  min_observations: int
    :param nruns: only used if assume normal is false
    :type  nruns: int
    :param relative: if relative==True, then the values will be returned
            as distances below and above the mean, respectively, rather than the
            absolute values. In	this case, the interval is mean-ret_val[0] to
            mean+ret_val[1]. This is more useful in many situations because it
            corresponds with the sem() and std() functions.
    :type: relative: boolean
    
    :return: results of type SimpleTestStatistics
    :rtype: SimpleTestStatistics
    """
    # Check if data was provided and it has correct format
    if x is None or y is None:
        raise ValueError('Please provide two non-None samples.')
    if not isinstance(x, pd.Series) and not isinstance(x, np.ndarray) and not isinstance(x, list):
        raise TypeError('Please provide samples of type Series or list.')
    if type(x) != type(y):
        raise TypeError('Please provide samples of the same type.')

    percentiles = [alpha * 100 / 2, 100 - alpha * 100 / 2]

    # Coercing missing values to right format
    _x = np.array(x, dtype=float)
    _y = np.array(y, dtype=float)

    x_nan = np.isnan(_x).sum()
    y_nan = np.isnan(_y).sum()
    if x_nan > 0:
        warnings.warn('Discarding ' + str(x_nan) + ' NaN(s) in the x array!')
        logger.warning('Discarding ' + str(x_nan) + ' NaN(s) in the x array!')
    if y_nan > 0:
        warnings.warn('Discarding ' + str(y_nan) + ' NaN(s) in the y array!')
        logger.warning('Discarding ' + str(x_nan) + ' NaN(s) in the x array!')

    ss_x = sample_size(_x)
    ss_y = sample_size(_y)

    # Checking if enough observations are left after dropping NaNs
    if min(ss_x, ss_y) < min_observations:
        # Set mean to nan
        mu = np.nan
        # Create nan dictionary
        c_i = dict(list(zip(percentiles, np.empty(len(percentiles)) * np.nan)))
    else:
        # Computing the mean
        mu = _delta_mean(_x, _y)
        # Computing the confidence intervals
        if assume_normal:
            logger.info("The distribution of two samples is assumed normal. "
                        "Performing the sample difference distribution calculation.")
            c_i = normal_sample_difference(x=_x, y=_y, percentiles=percentiles, relative=relative)
        else:
            logger.info("The distribution of two samples is not normal. Performing the bootstrap.")
            c_i, _ = bootstrap(x=_x, y=_y, percentiles=percentiles, nruns=nruns, relative=relative)

    treatment_statistics = SampleStatistics(ss_x, float(np.nanmean(_x)), float(np.nanvar(_x)))
    control_statistics   = SampleStatistics(ss_y, float(np.nanmean(_y)), float(np.nanvar(_y)))
    variant_statistics   = BaseTestStatistics(control_statistics, treatment_statistics)
    p_value              = compute_p_value_from_samples(_x, _y)
    statistical_power    = compute_statistical_power_from_samples(_x, _y, alpha)

    logger.info("Delta calculation finished!")
    return SimpleTestStatistics(variant_statistics.control_statistics,
                                variant_statistics.treatment_statistics,
                                float(mu), c_i, p_value, statistical_power)


def sample_size(x):
    """ Calculates valid sample size given the data.

    :param x: sample to calculate the sample size
    :type  x: pd.Series or list (array-like)

    :return: sample size of the sample excluding nans
    :rtype: int
    """
    # cast into a dummy numpy array to infer the dtype
    x_as_array = np.array(x)

    if np.issubdtype(x_as_array.dtype, np.number):
        _x = np.array(x, dtype=float)
        x_nan = np.isnan(_x).sum()
    # assuming categorical sample
    elif isinstance(x, pd.core.series.Series):
        x_nan = x.str.contains('NA').sum()
    else:
        x_nan = list(x).count('NA')

    return int(len(x) - x_nan)


def estimate_sample_size(x, mde, r, alpha=0.05, beta=0.2):
    """
    Estimates sample size based on sample mean and variance given MDE (Minimum Detectable effect), 
    number of variants and variant split ratio
    
    :param x: sample to base estimation on
    :type  x: pd.Series or pd.DataFrame
    :param mde: minimum detectable effect
    :type  mde: float
    :param r: variant split ratio
    :type  r: float
    :param alpha: significance level
    :type  alpha: float
    :param beta: type II error
    :type  beta: float
    
    :return: estimated sample size
    :rtype: float or pd.Series
    """
    if not isinstance(x, pd.Series) and not isinstance(x, pd.DataFrame):
        raise TypeError("Sample x needs to be either Series or DataFrame.")

    if r <= 0:
        raise ValueError("Variant split ratio needs to be higher than 0.")

    ppf = stats.norm.ppf
    c1 = (ppf(1.0 - alpha/2.0) - ppf(beta))**2
    c2 = (1.0 + r) * c1 * (1.0 + 1.0 / r)
    return c2 * x.var() / (mde * x.mean())**2


def bootstrap(x, y, func=_delta_mean, nruns=10000, percentiles=[2.5, 97.5],
              min_observations=20, return_bootstraps=False, relative=False):
    """ Bootstraps the Confidence Intervals for a particular function comparing two samples. 
    NaNs are ignored (discarded before calculation).

    :param x: sample of the treatment group
    :type  x: pd.Series or list (array-like)
    :param y: sample of the control group
    :type  y: pd.Series or list (array-like)
    :param func: function of which the distribution is to be computed.
                 The default comparison metric is the difference of means. 
                 For bootstraping correlation: func=lambda x,y: np.stats.pearsonr(x,y)[0].
    :type  func: function
    :param nruns: number of bootstrap runs to perform
    :type  nruns: int
    :param percentiles: The values corresponding to the given percentiles are returned. 
                        The default percentiles (2.5% and 97.5%) correspond to an alpha of 0.05.
    :type  percentiles: list
    :param min_observations: minimum number of observations necessary
    :type  min_observations: int
    :param return_bootstraps: If this variable is set the bootstrap sets are returned,
                              otherwise the first return value is empty.
    :type  return_bootstraps: bool
    :param relative: if relative==True, then the values will be returned as distances below and above the mean, 
                     respectively, rather than the absolute values. 
                     In this case, the interval is mean-ret_val[0] to mean+ret_val[1]. 
                     This is more useful in many situations because it corresponds with the sem() and std() functions.
    :type  relative: bool
    
    :return (c_val, bootstraps): c_val is a dict which contains percentile levels (index) and values
                                 bootstraps is a np.array containing the bootstrapping results per run
    :rtype: tuple
    """
    # Checking if data was provided
    if x is None or y is None:
        raise ValueError('Please provide two non-None samples.')

    # Transform data to appropriate format
    _x = np.array(x, dtype=float)
    _y = np.array(y, dtype=float)
    ss_x = _x.size - np.isnan(_x).sum()
    ss_y = _y.size - np.isnan(_y).sum()

    # Checking if enough observations are left after dropping NaNs
    if min(ss_x, ss_y) < min_observations:
        # Create nan percentile dictionary
        c_val = dict(list(zip(percentiles, np.empty(len(percentiles)) * np.nan)))
        return (c_val, None)
    else:
        # Initializing bootstraps array and random sampling for each run
        bootstraps = np.ones(nruns) * np.nan
        for run in range(nruns):
            # Randomly choose values from _x and _y with replacement
            xp = _x[np.random.randint(0, len(_x), size=(len(_x),))]
            yp = _y[np.random.randint(0, len(_y), size=(len(_y),))]
            # Application of the given function to the bootstraps
            bootstraps[run] = func(xp, yp)
        # If relative is set subtract mean from bootstraps
        if relative:
            bootstraps -= np.nanmean(bootstraps)
        # Confidence values per given percentile as dictionary
        c_val = dict(list(zip(percentiles, np.percentile(bootstraps, q=percentiles))))
        return (c_val, None) if not return_bootstraps else (c_val, bootstraps)


def pooled_std(std1, n1, std2, n2):
    """ Returns the pooled estimate of standard deviation. 

    For further information visit:
        http://sphweb.bumc.bu.edu/otlt/MPH-Modules/BS/BS704_Confidence_Intervals/BS704_Confidence_Intervals5.html

    :param std1: standard deviation of first sample
    :type  std1: float
    :param n1: size of first sample
    :type  n1: int
    :param std2: standard deviation of second sample
    :type  std2: float
    :param n2: size of second sample
    :type  n2: int

    :return: pooled standard deviation
    :type: float
    """
    if not (0.5 < (std1 ** 2) / (std2 ** 2) < 2.):
        warnings.warn('Sample variances differ too much to assume that population variances are equal.')
        logger.warning('Sample variances differ too much to assume that population variances are equal.')

    return np.sqrt(((n1 - 1) * std1 ** 2 + (n2 - 1) * std2 ** 2) / (n1 + n2 - 2))


def normal_sample_difference(x, y, percentiles=[2.5, 97.5], relative=False):
    """ Calculates the difference distribution of two normal distributions given by their samples.

    Computation is done in form of treatment minus control. 
    It is assumed that the standard deviations of both distributions do not differ too much.

    :param x: sample of a treatment group
    :type  x: pd.Series or list (array-like)
    :param y: sample of a control group
    :type  x: pd.Series or list (array-like)
    :param percentiles: list of percentile values to compute
    :type  percentiles: list
    :param relative: If relative==True, then the values will be returned
                     as distances below and above the mean, respectively, rather than the
                     absolute values. In this case, the interval is mean-ret_val[0] to
                     mean+ret_val[1]. This is more useful in many situations because it
                     corresponds with the sem() and std() functions.
    :type relative: bool
    
    :return: percentiles and corresponding values
    :rtype: dict
    """
    # Coerce data to right format
    _x = np.array(x, dtype=float)
    _x = _x[~np.isnan(_x)]
    _y = np.array(y, dtype=float)
    _y = _y[~np.isnan(_y)]

    # Calculate statistics
    mean1 = np.mean(_x)
    mean2 = np.mean(_y)
    std1 = np.std(_x)
    std2 = np.std(_y)
    n1 = len(_x)
    n2 = len(_y)

    # Push calculation to normal difference function
    return normal_difference(mean1=mean1, std1=std1, n1=n1,
                             mean2=mean2, std2=std2, n2=n2,
                             percentiles=percentiles, relative=relative)


def normal_difference(mean1, std1, n1, mean2, std2, n2, percentiles=[2.5, 97.5], relative=False):
    """ Calculates the difference distribution of two normal distributions.
    Computation is done in form of treatment minus control. It is assumed that
    the standard deviations of both distributions do not differ too much.

    For further information visit:
        http://sphweb.bumc.bu.edu/otlt/MPH-Modules/BS/BS704_Confidence_Intervals/BS704_Confidence_Intervals5.html

    :param mean1: mean value of the treatment distribution
    :type  mean1: float
    :param std1: standard deviation of the treatment distribution
    :type  std1: float
    :param n1: number of samples of the treatment distribution
    :type  n1: int
    :param mean2: mean value of the control distribution
    :type  mean2: float
    :param std2: standard deviation of the control distribution
    :type  std2: float
    :param n2: number of samples of the control distribution
    :type  n2: int
    :param percentiles: list of percentile values to compute
    :type  percentiles: list
    :param relative: If relative==True, then the values will be returned
                     as distances below and above the mean, respectively, rather than the
                     absolute values. In this case, the interval is mean-ret_val[0] to
                     mean+ret_val[1]. This is more useful in many situations because it
                     corresponds with the sem() and std() functions.
    :type relative: bool
    
    :return: percentiles and corresponding values
    :rtype: dict
    """
    # Compute combined parameters from individual parameters
    mean = mean1 - mean2
    std = pooled_std(std1, n1, std2, n2)
    # Computing standard error
    st_error = std * np.sqrt(1. / n1 + 1. / n2)
    # Computing degrees of freedom
    d_free = n1 + n2 - 2

    # Mapping percentiles via standard error
    if relative:
        return dict([(round(p, 5), stats.t.ppf(p / 100.0, df=d_free) * st_error) for p in percentiles])
    else:
        return dict([(round(p, 5), mean + stats.t.ppf(p / 100.0, df=d_free) * st_error) for p in percentiles])


def compute_statistical_power_from_samples(x, y, alpha=0.05):
    """ Compute statistical power given data samples of control and treatment.

    :param x: samples of a treatment group
    :type  x: pd.Series or array-like
    :param y: samples of a control group
    :type  y: pd.Series or array-like
    :param alpha: Type I error (false positive rate)
    :type  alpha: float
    
    :return: statistical power---the probability of a test to detect an effect if the effect actually exists
    :rtype: float
    """
    z_1_minus_alpha = stats.norm.ppf(1 - alpha/2.)
    _x = np.array(x, dtype=float)
    _x = _x[~np.isnan(_x)]
    _y = np.array(y, dtype=float)
    _y = _y[~np.isnan(_y)]

    mean1 = np.mean(_x)
    mean2 = np.mean(_y)
    std1 = np.std(_x)
    std2 = np.std(_y)
    n1 = len(_x)
    n2 = len(_y)
    return compute_statistical_power(mean1, std1, n1, mean2, std2, n2, z_1_minus_alpha)


def compute_statistical_power(mean1, std1, n1, mean2, std2, n2, z_1_minus_alpha):
    """ Compute statistical power given statistics of control and treatment.
    
    :param mean1: mean value of the treatment distribution
    :type  mean1: float
    :param std1: standard deviation of the treatment distribution
    :type  std1: float
    :param n1: number of samples of the treatment distribution
    :type  n1: int
    :param mean2: mean value of the control distribution
    :type  mean2: float
    :param std2: standard deviation of the control distribution
    :type  std2: float
    :param n2: number of samples of the control distribution
    :type  n2: int
    :param z_1_minus_alpha: critical value for significance level alpha. That is, z-value for 1-alpha.
    :type  z_1_minus_alpha: float
    
    :return: statistical power---the probability of a test to detect an effect if the effect actually exists
    :rtype: float
    """
    effect_size = mean1 - mean2
    std = pooled_std(std1, n1, std2, n2)
    if std <= 0.0:
        raise ValueError('Zero pooled std in compute_statistical_power')

    tmp = (n1 * n2 * effect_size**2) / ((n1 + n2) * std**2)
    z_beta = z_1_minus_alpha - np.sqrt(tmp)
    beta = stats.norm.cdf(z_beta)
    power = 1 - beta
    return power


def compute_p_value_from_samples(x, y):
    """ Calculates two-tailed p value for statistical Student's T-test based on pooled standard deviation.

    :param x: samples of a treatment group
    :type  x: pd.Series or array-like
    :param y: samples of a control group
    :type  y: pd.Series or array-like

    :return: two-tailed p-value 
    :rtype: float
    """
    if x is None or y is None:
        raise ValueError('Please provide two non-empty samples to compute p-values.')

    _x = np.array(x, dtype=float)
    _x = _x[~np.isnan(_x)]
    _y = np.array(y, dtype=float)
    _y = _y[~np.isnan(_y)]

    mean1 = np.mean(_x)
    mean2 = np.mean(_y)
    std1 = np.std(_x)
    std2 = np.std(_y)
    n1 = len(_x)
    n2 = len(_y)
    return compute_p_value(mean1, std1, n1, mean2, std2, n2)


def compute_p_value(mean1, std1, n1, mean2, std2, n2):
    """ Compute two-tailed p value for statistical Student's T-test given statistics of control and treatment.
    
    :param mean1: mean value of the treatment distribution
    :type  mean1: float
    :param std1: standard deviation of the treatment distribution
    :type  std1: float
    :param n1: number of samples of the treatment distribution
    :type  n1: int
    :param mean2: mean value of the control distribution
    :type  mean2: float
    :param std2: standard deviation of the control distribution
    :type  std2: float
    :param n2: number of samples of the control distribution
    :type  n2: int
    
    :return: two-tailed p-value 
    :rtype: float
    """
    mean_diff = mean1 - mean2
    std       = pooled_std(std1, n1, std2, n2)
    st_error  = std * np.sqrt(1. / n1 + 1. / n2)
    d_free    = n1 + n2 - 2
    t         = mean_diff / st_error
    p         = stats.t.cdf(-abs(t), df=d_free) * 2
    return p



import logging
import warnings

import numpy as np
import pandas as pd
import copy

import expan.core.early_stopping as es
import expan.core.statistics as statx
import expan.core.correction as correction
from expan.core.statistical_test import *
from expan.core.results import StatisticalTestResult, MultipleTestSuiteResult, CombinedTestStatistics

warnings.simplefilter('always', UserWarning)
warnings.filterwarnings("ignore", category=FutureWarning)
logger = logging.getLogger(__name__)


class Experiment(object):
    """ Class which adds the analysis functions to experimental data. """
    def __init__(self, metadata):
        """
        Constructor of the experiment object.
        
        :param metadata: additional information about the experiment. (e.g. primary KPI, source, etc)
        :type  metadata: dict
        """
        self.metadata     = metadata
        self.worker_table = {
            'fixed_horizon': statx.make_delta,
            'group_sequential': es.make_group_sequential,
            'bayes_factor': es.make_bayes_factor,
            'bayes_precision': es.make_bayes_precision
        }

    def __str__(self):
        return 'Performing "{:s}" experiment.'.format(self.metadata['experiment'])


    def analyze_statistical_test(self, test, test_method='fixed_horizon', include_data=False, **worker_args):
        """ Runs delta analysis on one statistical test and returns statistical results.
        
        :param test: a statistical test to run
        :type  test: StatisticalTest
        :param test_method: analysis method to perform. 
                           It can be 'fixed_horizon', 'group_sequential', 'bayes_factor' or 'bayes_precision'.
        :type  test_method: str
        :param include_data: True if test results should include data, False - if no data should be included
        :type  include_data: bool
        :param worker_args: additional arguments for the analysis method

        :return: statistical result of the test
        :rtype: StatisticalTestResult
        """
        if not isinstance(test, StatisticalTest):
            raise TypeError("Statistical test should be of type StatisticalTest.")

        if 'entity' not in test.data.columns:
            raise KeyError("There is no 'entity' column in the data.")
        if test.variants.variant_column_name not in test.data.columns:
            raise KeyError("There is no '{}' column in the data.".format(test.variants.variant_column_name))
        if test.variants.treatment_name not in pd.unique(test.data[test.variants.variant_column_name]):
            raise KeyError("There is no treatment with the name '{}' in the data.".format(test.variants.treatment_name))
        if test.variants.control_name not in pd.unique(test.data[test.variants.variant_column_name]):
            raise KeyError("There is no control with the name '{}' in the data.".format(test.variants.control_name))

        for feature in test.features:
            if feature.column_name not in test.data.columns:
                raise KeyError("Feature name '{}' does not exist in the data.".format(feature.column_name))

        if type(test.kpi) is KPI and (test.kpi.name not in test.data.columns):
            raise KeyError("There is no column of name '{}' in the data.".format(test.kpi.name))
        if type(test.kpi) is DerivedKPI:
            if type(test.kpi.numerator) is not str or test.kpi.numerator not in test.data.columns:
                raise KeyError("Numerator '{}' of the derived KPI does not exist in the data.".format(test.kpi.numerator))
            if type(test.kpi.denominator) is not str or test.kpi.denominator not in test.data.columns:
                raise KeyError("Denominator '{}' of the derived KPI does not exist in the data.".format(test.kpi.denominator))
            test.kpi.make_derived_kpi(test.data)

        logger.info("One analysis with kpi '{}', control variant '{}', treatment variant '{}' and features [{}] "
                    "has just started".format(test.kpi, test.variants.control_name,
                                              test.variants.treatment_name,
                                              [(feature.column_name, feature.column_value) for feature in test.features]))

        if test_method not in self.worker_table:
            raise NotImplementedError("Test method '{}' is not implemented.".format(test_method))
        worker = self.worker_table[test_method](**worker_args)

        # create test result object with empty result first
        test_result = StatisticalTestResult(test, None)

        data_for_analysis = test.data
        for feature in test.features:
            data_for_analysis = feature.apply_to_data(data_for_analysis)

        if not self._is_valid_for_analysis(data_for_analysis, test):
            logger.warning("Data is not valid for the analysis!")
            return test_result
        if data_for_analysis.entity.duplicated().any():
            raise ValueError('Entities in data should be unique.')

        # get control and treatment values for the kpi
        control          = test.variants.get_variant(data_for_analysis, test.variants.control_name)[test.kpi.name]
        control_weight   = self._get_weights(data_for_analysis, test, test.variants.control_name)
        control_data     = control * control_weight

        treatment        = test.variants.get_variant(data_for_analysis, test.variants.treatment_name)[test.kpi.name]
        treatment_weight = self._get_weights(data_for_analysis, test, test.variants.treatment_name)
        treatment_data   = treatment * treatment_weight

        # run the test method
        test_statistics = worker(x=treatment_data, y=control_data)
        test_result.result = test_statistics

        # remove data from the result test metadata
        if not include_data:
            del test.data
        return test_result


    def analyze_statistical_test_suite(self, test_suite, test_method='fixed_horizon', **worker_args):
        """ Runs delta analysis on a set of tests and returns statistical results for each statistical test in the suite.
        
        :param test_suite: a suite of statistical test to run
        :type  test_suite: StatisticalTestSuite
        :param test_method: analysis method to perform. 
                           It can be 'fixed_horizon', 'group_sequential', 'bayes_factor' or 'bayes_precision'.
        :type  test_method: str
        :param worker_args: additional arguments for the analysis method (see signatures of corresponding methods)

        :return: statistical result of the test suite
        :rtype: MultipleTestSuiteResult
        """
        if not isinstance(test_suite, StatisticalTestSuite):
            raise TypeError("Test suite should be of type StatisticalTestSuite.")

        if test_method not in ['fixed_horizon', 'group_sequential']:
            test_suite.correction_method = CorrectionMethod.NONE
        requires_correction = test_suite.correction_method is not CorrectionMethod.NONE

        # look up table for correction method
        correction_table = {
            CorrectionMethod.BONFERRONI: correction.bonferroni,
            CorrectionMethod.BH: correction.benjamini_hochberg
        }

        logger.info("Statistical test suite analysis with {} tests, testmethod {}, correction method {} "
                    "has just started".format(len(test_suite.tests), test_method, test_suite.correction_method))

        # test_suite_result hold statistical results from all statistical tests
        test_suite_result = MultipleTestSuiteResult([], test_suite.correction_method)
        for test in test_suite.tests:
            original_analysis = self.analyze_statistical_test(test, test_method, True, **worker_args)
            combined_result = CombinedTestStatistics(original_analysis.result, original_analysis.result)
            original_analysis.result = combined_result
            test_suite_result.results.append(original_analysis)

        # if correction is needed, get p values, do correction on alpha, and run the same analysis for new alpha
        if requires_correction:
            original_alpha    = worker_args.get('alpha', 0.05)
            original_p_values = [item.result.original_test_statistics.p for item in test_suite_result.results
                                 if item.result.original_test_statistics is not None]
            corrected_alpha   = correction_table[test_suite.correction_method](original_alpha, original_p_values)
            new_worker_args   = copy.deepcopy(worker_args)
            new_worker_args['alpha'] = corrected_alpha

            for test_index, test_item in enumerate(test_suite_result.results):
                if test_item.result.original_test_statistics:  # result can be None if not enough entities
                    original_analysis = test_suite_result.results[test_index]
                    corrected_analysis = self.analyze_statistical_test(test_item.test, test_method,
                                                                       True, **new_worker_args)
                    combined_result = CombinedTestStatistics(original_analysis.result.original_test_statistics,
                                                             corrected_analysis.result)
                    original_analysis.result = combined_result

        logger.info("Statistical test suite analysis with {} tests, testmethod {}, correction method {} "
                    "has finished".format(len(test_suite.tests), test_method, test_suite.correction_method))

        # remove data from the results
        for test_result in test_suite_result.results:
            del test_result.test.data
        return test_suite_result


    def outlier_filter(self, data, kpis, percentile=99.0, threshold_type='upper'):
        """ Method that filters out entities whose KPIs exceed the value at a given percentile.
        If any of the KPIs exceeds its threshold the entity is filtered out.
        
        :param kpis: list of KPI names
        :type  kpis: list[str]
        :param percentile: percentile considered as filtering threshold
        :type  percentile: float
        :param threshold_type: type of threshold used ('lower' or 'upper')
        :type  threshold_type: str

        :return: Will return data with filtered outliers.
        """
        # check if provided KPIs are present in the data
        for kpi in kpis:
            if kpi not in data.columns:
                raise KeyError(kpi + ' identifier not present in dataframe columns!')
        # check if provided percentile is valid
        if 0.0 < percentile <= 100.0 is False:
            raise ValueError("Percentile value needs to be between 0.0 and 100.0!")
        # check if provided filtering kind is valid
        if threshold_type not in ['upper', 'lower']:
            raise ValueError("Threshold type needs to be either 'upper' or 'lower'!")

        # run quantile filtering
        flags = self._quantile_filtering(data=data, kpis=kpis, percentile=percentile, threshold_type=threshold_type)
        # log which columns were filtered and how many entities were filtered out
        self.metadata['filtered_columns'] = kpis
        self.metadata['filtered_entities_number'] = len(flags[flags == True])
        self.metadata['filtered_threshold_kind'] = threshold_type
        # throw warning if too many entities have been filtered out
        if (len(flags[flags == True]) / float(len(data))) > 0.02:
            warnings.warn('More than 2% of entities have been filtered out, consider adjusting the percentile value.')
            logger.warning('More than 2% of entities have been filtered out, consider adjusting the percentile value.')
        return data[flags == False]


    # ----- below are helper methods ----- #
    def _is_valid_for_analysis(self, data, test):
        """ Check whether the quality of data is good enough to perform analysis. Invalid cases can be:
        1. there is no data
        2. the data does not contain all the variants to perform analysis
        
        :param data: data frame for which a check for validity will be made
        :type  data: DataFrame
        :param test: a statistical test for control name and treatment name
        :type  test: StatisticalTest
        
        :return True if data is valid for analysis and False if not
        :rtype: bool 
        """
        if len(data[data[test.variants.variant_column_name] == test.variants.control_name]) <= 1:
            logger.warning("Control group only contains 1 or 0 entities.")
            return False
        if len(data[data[test.variants.variant_column_name] == test.variants.treatment_name]) <= 1:
            logger.warning("Treatment group only contains 1 or 0 entities.")
        return True


    def _get_weights(self, data, test, variant_name):
        """ Perform the re-weighting trick on the selected derived kpi
        See http://expan.readthedocs.io/en/latest/glossary.html#per-entity-ratio-vs-ratio-of-totals
        
        :param data: data subset by derived kpi and variant, for which the re-weighting trick should be done
        :type  data: pd.DataFrame
        :param test: statistical test, provides the derived kpi denominator for the re-weighting trick
        :type  test: StatisticalTest
        :param variant_name: variant name for data subset by kpi and variant
        :type  variant_name: str
        
        :return returns re-weighted kpi values of type pd.Series
        :rtype: pd.Series
        """
        if type(test.kpi) is not DerivedKPI:
            return 1.0

        x = test.variants.get_variant(data, variant_name)[test.kpi.denominator]
        number_of_zeros_and_nans     = sum(x == 0) + np.isnan(x).sum()
        number_of_non_zeros_and_nans = len(x) - number_of_zeros_and_nans
        return number_of_non_zeros_and_nans / np.nansum(x) * x


    def _quantile_filtering(self, data, kpis, percentile, threshold_type):
        """ Make the filtering based on the given quantile level. 
        Filtering is performed for each kpi independently.
        
        :param kpis: the kpis to perform filtering
        :type  kpis: list[str]
        :param percentile: percentile considered as filtering threshold
        :type  percentile: float
        :param threshold_type: type of threshold used ('lower' or 'upper')
        :type  threshold_type: str

        :return: boolean values indicating whether the row should be filtered
        :rtype: pd.Series
        """
        method_table = {'upper': lambda x: x > threshold, 'lower': lambda x: x <= threshold}
        flags = pd.Series(data=[False]*len(data))
        for column in data[kpis].columns:
            threshold = np.percentile(data[column], percentile)
            flags = flags | data[column].apply(method_table[threshold_type])
        return flags



from enum import Enum

import pandas as pd
from expan.core.util import JsonSerializable


class StatisticalTest(JsonSerializable):
    """ This class describes what has to be tested against what and represent a unit of statistical testing.
    
    :param data: data for statistical test
    :type  data: DataFrame
    :param kpi: the kpi to perform on
    :type  kpi: KPI or its subclass
    :param features: list of features used for subgroups
    :type  features: list[FeatureFilter]
    :param variants: variant column name and their values
    :type  variants: Variants
    """
    def __init__(self, data, kpi, features, variants):
        if not isinstance(data, pd.DataFrame):
            raise ValueError("Please, provide data for the statistical test in form of a data frame.")
        if not isinstance(features, list):
            raise TypeError("Features should be a list.")
        if not all(isinstance(n, FeatureFilter) for n in features):
            raise TypeError("Some features are not of the type FeatureFilter.")
        self.data     = data
        self.kpi      = kpi
        self.features = features
        self.variants = variants


class KPI(JsonSerializable):
    """ This class represents a basic kpi.
    :param name: name of the kpi
    :type  name: str
    """
    def __init__(self, name):
        self.name = name


class DerivedKPI(KPI):
    """ This class represents a derived KPI which is a ratio of two columns. 
    Names of the the two columns are passed as numerator and denominator.
    
    :param name: name of the kpi
    :type  name: str
    :param numerator: the numerator for the derived KPI
    :type  numerator: str
    :param denominator: the denominator for the derived KPI
    :type  denominator: str
    """
    def __init__(self, name, numerator, denominator):
        super(DerivedKPI, self).__init__(name)
        self.numerator = numerator
        self.denominator = denominator

    def make_derived_kpi(self, data):
        """ Create the derived kpi column if it is not yet created. """
        if self.name not in data.columns:
            data.loc[:, self.name] = (data[self.numerator]/data[self.denominator]).astype("float64")


class CorrectionMethod(Enum):
    """ Correction methods."""
    NONE       = 1   # no correction
    BONFERRONI = 2   # Bonferroni correction. Used to correct false positive rate.
    BH         = 3   # Benjamini-Hochberg procedure. Used to correct false discovery rate.


class StatisticalTestSuite(JsonSerializable):
    """ This class consists of a number of tests plus choice of the correction method.
    
    :param tests: list of statistical tests in the suite
    :type  tests: list[StatisticalTest]
    :param correction_method: method used for multiple testing correction
    :type  correction_method: CorrectionMethod
    """
    def __init__(self, tests, correction_method=CorrectionMethod.NONE):
        if len(tests) is 1:
            correction_method = CorrectionMethod.NONE
        self.tests = tests
        self.correction_method = correction_method

    @property
    def size(self):
        return len(self.tests)


class FeatureFilter(JsonSerializable):
    """ This class represents a filter, restricting a DataFrame to rows with column_value in column_name.
    
    It can be used to specify subgroup conditions.
    :param column_name: name of the column to perform filter on
    :type  column_name: str
    :param column_value: value of the column to perform filter on
    :type  column_value: str
    """
    def __init__(self, column_name, column_value):
        self.column_name  = column_name
        self.column_value = column_value

    def apply_to_data(self, data):
        return data[data[self.column_name] == self.column_value]


class Variants(JsonSerializable):
    """ This class represents information of variants.
    
    :param variant_column_name: name of the column that represents variant
    :type  variant_column_name: str
    :param control_name: value of the variant that represents control group
    :type  control_name: str
    :param treatment_name: value of the variant that represents control group
    :type  treatment_name: str
    """
    def __init__(self, variant_column_name, control_name, treatment_name):
        self.variant_column_name = variant_column_name
        self.control_name        = control_name
        self.treatment_name      = treatment_name

    def get_variant(self, data, variant_name):
        result = data[data[self.variant_column_name] == variant_name]
        if not isinstance(result, pd.DataFrame):
            result = pd.DataFrame([result])
        return result



from expan.core.statistical_test import CorrectionMethod
from expan.core.util import JsonSerializable


# --------- Below are the data structure of statistics --------- #
class BaseTestStatistics(JsonSerializable):
    """ Holds only statistics for the control and treatment group. 
    
    :param control_statistics: statistics within the control group
    :type  control_statistics: SampleStatistics
    :param treatment_statistics: statistics within the treatment group
    :type  treatment_statistics: SampleStatistics
    """
    def __init__(self, control_statistics, treatment_statistics):
        self.control_statistics   = control_statistics
        self.treatment_statistics = treatment_statistics


class SampleStatistics(JsonSerializable):
    """ This class holds sample size, mean and variance.
    
    :param sample_size: samples size of the control or treatment group
    :type  sample_size: int
    :param mean: mean of the control or treatment group
    :type  mean: float
    :param variance: variance of the control or treatment group
    :type variance: float
    """
    def __init__(self, sample_size, mean, variance):
        self.sample_size = sample_size
        self.mean        = mean
        self.variance    = variance


class SimpleTestStatistics(BaseTestStatistics):
    """ Additionally to BaseTestStatistics, holds delta, confidence interval, statistical power, and p value.
    
    :param control_statistics: sample size, mean, variance for the control group
    :type  control_statistics: SampleStatistics
    :param treatment_statistics: sample size, mean, variance for the treatment group
    :type  treatment_statistics: SampleStatistics
    :param delta: delta (relative or absolute difference between control and treatment, uplift)
    :type  delta: float
    :param p: p value
    :type  p: float
    :param statistical_power: statistical power value
    :type  statistical_power: float
    :param ci: a dict where keys are percentiles and values are the corresponding value for the statistic.
    :type  ci: dict
    """
    def __init__(self, control_statistics, treatment_statistics, delta, ci, p, statistical_power):
        super(SimpleTestStatistics, self).__init__(control_statistics, treatment_statistics)
        self.delta               = delta
        self.p                   = p
        self.statistical_power   = statistical_power
        self.confidence_interval = [{'percentile': p, 'value': v} for (p, v) in ci.items()]


class EarlyStoppingTestStatistics(SimpleTestStatistics):
    """ Additionally to SimpleTestStatistics, holds boolean flag for early stopping.
    
    :param control_statistics: sample size, mean, variance for the control group
    :type  control_statistics: SampleStatistics
    :param treatment_statistics: sample size, mean, variance for the treatment group
    :type  treatment_statistics: SampleStatistics
    :type  delta: float
    :type  p: float
    :type  statistical_power: float
    :param ci: a dict where keys are percentiles and values are the corresponding value for the statistic.
    :type  ci: dict
    :param stop: early-stopping flag
    :type  stop: bool
    """
    def __init__(self, control_statistics, treatment_statistics, delta, ci, p, statistical_power, stop):
        super(EarlyStoppingTestStatistics, self).__init__(control_statistics, treatment_statistics, delta, ci, p, statistical_power)
        self.stop = stop


class CombinedTestStatistics(JsonSerializable):
    """ Holds original and corrected statistics. This class should be used to hold statistics for multiple testing.
    original_test_statistics and corrected_test_statistics should have the same type.
    In case there is no correction specified, corrected_test_statistics == original_test_statistics.
    
    :param original_test_statistics: test result before correction
    :type  original_test_statistics: SimpleTestStatistics or EarlyStoppingTestStatistics
    :param corrected_test_statistics: test result after correction or same as original_test_statistics if no correction
    :type  corrected_test_statistics: SimpleTestStatistics or EarlyStoppingTestStatistics
    """
    def __init__(self, original_test_statistics, corrected_test_statistics):
        type1 = type(original_test_statistics)
        type2 = type(corrected_test_statistics)
        if type1 != type2:
            raise RuntimeError("Type mismatch for type " + str(type1) + " and " + str(type2))
        self.original_test_statistics  = original_test_statistics
        self.corrected_test_statistics = corrected_test_statistics


# --------- Below are the data structure of test results --------- #
class StatisticalTestResult(JsonSerializable):
    """ This class holds the results of a single statistical test.
    
    :param test: information about the statistical test
    :type  test: StatisticalTest
    :param result: result of this statistical test
    :type  result: CombinedTestStatistics
    """
    def __init__(self, test, result):
        self.test   = test
        self.result = result


class MultipleTestSuiteResult(JsonSerializable):
    """ This class holds the results of a MultipleTestSuite.
    
    :param results: test results for all statistical testing unit
    :type  results: list[StatisticalTestResult]
    :param correction_method: method used for multiple testing correction
    :type  correction_method: CorrectionMethod
    """
    def __init__(self, results, correction_method=CorrectionMethod.NONE):
        self.results = results
        self.correction_method = correction_method



import unittest
import warnings

import numpy as np

from expan.core.results import CombinedTestStatistics
from expan.core.statistical_test import *
from expan.core.experiment import Experiment
from expan.core.util import generate_random_data, find_value_by_key_with_condition


class ExperimentTestCase(unittest.TestCase):
    """ Defines the setUp() and tearDown() functions for the statistics test cases."""

    def setUp(self):
        """ Load the needed data sets for all StatisticsTestCases and set the random
        seed so that randomized algorithms show deterministic behaviour."""
        np.random.seed(0)
        data, metadata = generate_random_data()

        self.data, self.metadata = data, metadata

        feature_has = FeatureFilter('feature', 'has')
        feature_non = FeatureFilter('feature', 'non')
        feature_one = FeatureFilter('feature', 'feature that only has one data point')

        # simple statistical test
        self.kpi = KPI('normal_same')
        self.variants = Variants('variant', 'B', 'A')
        self.nonsense_variants = Variants('variant', 'C', 'D')
        self.test_normal_same = StatisticalTest(self.data, self.kpi, [], self.variants)
        self.test_nonsense_variant = StatisticalTest(self.data, self.kpi, [], self.nonsense_variants)
        self.test_normal_same_feature_non = StatisticalTest(self.data, self.kpi, [feature_non], self.variants)
        self.test_normal_same_feature_has = StatisticalTest(self.data, self.kpi, [feature_has], self.variants)
        self.test_normal_same_feature_one = StatisticalTest(self.data, self.kpi, [feature_one], self.variants)

        # statistical test with derived kpi
        self.derived_kpi = DerivedKPI('derived_kpi_one', 'normal_same', 'normal_shifted')
        self.test_derived_kpi = StatisticalTest(self.data, self.derived_kpi, [], self.variants)

        # small dummy data frame
        data_dummy = np.array([['index', 'entity', 'variant', 'normal_same', 'normal_shifted'],
                               [0, 1, 'A', 2.0, 1.0], [1, 2, 'B', 3.0, 2.0],
                               [2, 3, 'A', 3.0, 2.0], [3, 4, 'B', 2.5, 1.0]])
        self.data_dummy_df = pd.DataFrame(data=data_dummy[1:, 1:],
                                          columns=data_dummy[0, 1:]).convert_objects(convert_numeric=True)

        data_dummy_with_nan = np.array([['index', 'entity', 'variant', 'normal_same', 'normal_shifted'],
                               [0, 1, 'A', 2.0, 1.0], [1, 2, 'B', 3.0, 2.0],
                               [2, 3, 'A', 3.0, 2.0], [3, 4, 'B', 2.5, 1.0],
                               [4, 5, 'A', 0.0, 0.0], [5, 6, 'B', None, None]])
        self.data_dummy_df_with_nan = pd.DataFrame(data=data_dummy_with_nan[1:, 1:],
                                                   columns=data_dummy_with_nan[0, 1:]).convert_objects(convert_numeric=True)

        # statistical test suite
        self.suite_with_one_test = StatisticalTestSuite([self.test_normal_same])
        self.suite_with_one_test_correction = StatisticalTestSuite([self.test_normal_same], CorrectionMethod.BH)
        self.suite_with_two_tests = StatisticalTestSuite([self.test_normal_same, self.test_derived_kpi],
                                                         CorrectionMethod.BONFERRONI)
        self.suite_with_one_subgroup = StatisticalTestSuite([self.test_normal_same_feature_one])
        self.suite_with_three_subgroups = StatisticalTestSuite([self.test_normal_same_feature_non,
                                                                self.test_normal_same_feature_has,
                                                                self.test_normal_same_feature_one],
                                                               CorrectionMethod.BH)

    def tearDown(self):
        """ Clean up after the test """
        pass

    def getExperiment(self):
        return Experiment(self.metadata)


class ExperimentClassTestCases(ExperimentTestCase):

    def test_Experiment_constructor(self):
        experiment = self.getExperiment()
        self.assertTrue(isinstance(experiment.metadata, dict))


class StatisticalTestTestCases(ExperimentTestCase):
    """ Test the method analyze_statistical_test. """

    def test_fixed_horizon(self):
        ndecimals = 5
        res = self.getExperiment().analyze_statistical_test(self.test_normal_same, test_method='fixed_horizon')

        self.assertAlmostEqual(res.result.delta, 0.033053, ndecimals)

        lower_bound_ci = find_value_by_key_with_condition(res.result.confidence_interval, 'percentile', 2.5, 'value')
        upper_bound_ci = find_value_by_key_with_condition(res.result.confidence_interval, 'percentile', 97.5, 'value')
        self.assertAlmostEqual(lower_bound_ci, -0.007135, ndecimals)
        self.assertAlmostEqual(upper_bound_ci, 0.073240, ndecimals)

        self.assertEqual(res.result.treatment_statistics.sample_size, 6108)
        self.assertEqual(res.result.control_statistics.sample_size,   3892)

        self.assertAlmostEqual(res.result.treatment_statistics.mean,  0.025219, ndecimals)
        self.assertAlmostEqual(res.result.control_statistics.mean,   -0.007833, ndecimals)

        self.assertAlmostEqual(res.result.statistical_power, 0.36401, ndecimals)
        self.assertEqual(res.test.kpi.name, 'normal_same')

    def test_group_sequential(self):
        ndecimals = 5
        res = self.getExperiment().analyze_statistical_test(self.test_normal_same, test_method='group_sequential')

        self.assertAlmostEqual(res.result.delta, 0.033053, ndecimals)

        lower_bound_ci = find_value_by_key_with_condition(res.result.confidence_interval, 'percentile', 2.5, 'value')
        upper_bound_ci = find_value_by_key_with_condition(res.result.confidence_interval, 'percentile', 97.5, 'value')
        self.assertAlmostEqual(lower_bound_ci, -0.007135, ndecimals)
        self.assertAlmostEqual(upper_bound_ci, 0.073240, ndecimals)

        self.assertEqual(res.result.treatment_statistics.sample_size, 6108)
        self.assertEqual(res.result.control_statistics.sample_size, 3892)

        self.assertAlmostEqual(res.result.treatment_statistics.mean, 0.025219, ndecimals)
        self.assertAlmostEqual(res.result.control_statistics.mean, -0.007833, ndecimals)

        self.assertAlmostEqual(res.result.statistical_power, 0.36401, ndecimals)
        self.assertEqual(res.result.stop, False)
        self.assertEqual(res.test.kpi.name, 'normal_same')

    def test_bayes_factor_delta(self):
        ndecimals = 5
        res = self.getExperiment().analyze_statistical_test(self.test_normal_same,
                                                            test_method='bayes_factor',
                                                            num_iters=2000)

        self.assertAlmostEqual(res.result.delta, 0.033053, ndecimals)

        lower_bound_ci = find_value_by_key_with_condition(res.result.confidence_interval, 'percentile', 2.5, 'value')
        upper_bound_ci = find_value_by_key_with_condition(res.result.confidence_interval, 'percentile', 97.5, 'value')
        self.assertAlmostEqual(lower_bound_ci, -0.00829, ndecimals)
        self.assertAlmostEqual(upper_bound_ci, 0.07127, ndecimals)

        self.assertEqual(res.result.treatment_statistics.sample_size, 6108)
        self.assertEqual(res.result.control_statistics.sample_size, 3892)

        self.assertAlmostEqual(res.result.treatment_statistics.mean, 0.025219, ndecimals)
        self.assertAlmostEqual(res.result.control_statistics.mean, -0.007833, ndecimals)

        self.assertEqual(res.result.stop, True)
        self.assertEqual(res.test.kpi.name, 'normal_same')

    def test_bayes_precision_delta(self):
        ndecimals = 5
        res = self.getExperiment().analyze_statistical_test(self.test_normal_same,
                                                            test_method='bayes_precision',
                                                            num_iters=2000)
        self.assertAlmostEqual(res.result.delta, 0.033053, ndecimals)

        lower_bound_ci = find_value_by_key_with_condition(res.result.confidence_interval, 'percentile', 2.5, 'value')
        upper_bound_ci = find_value_by_key_with_condition(res.result.confidence_interval, 'percentile', 97.5, 'value')
        self.assertAlmostEqual(lower_bound_ci, -0.00829, ndecimals)
        self.assertAlmostEqual(upper_bound_ci, 0.07127, ndecimals)

        self.assertEqual(res.result.treatment_statistics.sample_size, 6108)
        self.assertEqual(res.result.control_statistics.sample_size, 3892)

        self.assertAlmostEqual(res.result.treatment_statistics.mean, 0.025219, ndecimals)
        self.assertAlmostEqual(res.result.control_statistics.mean, -0.007833, ndecimals)

        self.assertEqual(res.result.stop, True)


class StatisticalTestSuiteTestCases(ExperimentTestCase):
    """ Test the method analyze_statistical_test_suite. """

    def test_one_test_in_suite(self):
        ndecimals = 5
        res = self.getExperiment().analyze_statistical_test_suite(self.suite_with_one_test)

        self.assertEqual(res.correction_method, CorrectionMethod.NONE)
        self.assertEqual(len(res.results), 1)
        simple_stat_res = res.results[0].result

        self.assertAlmostEqual(simple_stat_res.original_test_statistics.delta, 0.033053, ndecimals)
        lower_bound_ci = find_value_by_key_with_condition(simple_stat_res.original_test_statistics.confidence_interval,
                                                          'percentile', 2.5, 'value')
        upper_bound_ci = find_value_by_key_with_condition(simple_stat_res.original_test_statistics.confidence_interval,
                                                          'percentile', 97.5, 'value')
        self.assertAlmostEqual(lower_bound_ci, -0.007135, ndecimals)
        self.assertAlmostEqual(upper_bound_ci, 0.073240, ndecimals)

        self.assertEqual(simple_stat_res.original_test_statistics.treatment_statistics.sample_size, 6108)
        self.assertEqual(simple_stat_res.original_test_statistics.control_statistics.sample_size,   3892)

        self.assertAlmostEqual(simple_stat_res.original_test_statistics.treatment_statistics.mean,  0.025219, ndecimals)
        self.assertAlmostEqual(simple_stat_res.original_test_statistics.control_statistics.mean,   -0.007833, ndecimals)
        self.assertAlmostEqual(simple_stat_res.original_test_statistics.statistical_power, 0.36401, ndecimals)

        # corrected test statistics should be the same as original test statistics (checks some values)
        self.assertAlmostEqual(simple_stat_res.corrected_test_statistics.delta, 0.033053, ndecimals)
        self.assertAlmostEqual(simple_stat_res.corrected_test_statistics.statistical_power, 0.36401, ndecimals)
        self.assertEqual(simple_stat_res.corrected_test_statistics.treatment_statistics.sample_size, 6108)
        self.assertEqual(simple_stat_res.corrected_test_statistics.control_statistics.sample_size, 3892)

    def test_one_test_in_suite_with_wrong_correction(self):
        res = self.getExperiment().analyze_statistical_test_suite(self.suite_with_one_test_correction)
        self.assertEqual(res.correction_method, CorrectionMethod.NONE)
        self.assertEqual(len(res.results), 1)

    def test_two_tests_in_suite(self):
        ndecimals = 5

        res = self.getExperiment().analyze_statistical_test_suite(self.suite_with_two_tests)
        self.assertEqual(res.correction_method, CorrectionMethod.BONFERRONI)
        self.assertEqual(len(res.results), 2)

        res_normal_same = res.results[0]
        res_derived_kpi = res.results[1]

        self.assertEqual(res_normal_same.test.kpi.name, "normal_same")
        self.assertEqual(res_derived_kpi.test.kpi.name, "derived_kpi_one")
        self.assertTrue(isinstance(res_normal_same.result, CombinedTestStatistics))
        self.assertTrue(isinstance(res_derived_kpi.result, CombinedTestStatistics))

        power_normal_same_before = res_normal_same.result.original_test_statistics.statistical_power
        power_normal_same_corrected = res_normal_same.result.corrected_test_statistics.statistical_power
        self.assertAlmostEqual(power_normal_same_before, 0.36401, ndecimals)
        self.assertLess(power_normal_same_corrected, power_normal_same_before)

        power_derived_kpi_before = res_derived_kpi.result.original_test_statistics.statistical_power
        power_derived_kpi_corrected = res_derived_kpi.result.corrected_test_statistics.statistical_power
        self.assertAlmostEqual(power_derived_kpi_before, 0.3472, ndecimals)
        self.assertLess(power_derived_kpi_corrected, power_derived_kpi_before)

    def test_one_subgroup_in_suite(self):
        res = self.getExperiment().analyze_statistical_test_suite(self.suite_with_one_subgroup)
        self.assertEqual(res.correction_method, CorrectionMethod.NONE)
        self.assertEqual(len(res.results), 1)
        res_subgroup = res.results[0]

        self.assertEqual(len(res_subgroup.test.features), 1)
        self.assertEqual(res_subgroup.test.features[0].column_value, 'feature that only has one data point')
        self.assertIsNone(res_subgroup.result.original_test_statistics)
        self.assertIsNone(res_subgroup.result.corrected_test_statistics)

    def test_three_subgroup_in_suite(self):
        """ Two subgroups contain valid data. One subgroup contains only one entity.
        When there is no enough data in one subgroup, which means one subgroup might contain zero
        or only one variant. (e.g. "browser" = "hack name"), it should ignore this subgroup and continue
        doing analysis with other subgroups.
        """
        res = self.getExperiment().analyze_statistical_test_suite(self.suite_with_three_subgroups)
        self.assertEqual(res.correction_method, CorrectionMethod.BH)
        self.assertEqual(len(res.results), 3)

        res_subgroup_feature_non = res.results[0]
        res_subgroup_feature_has = res.results[1]
        res_subgroup_feature_one = res.results[2]

        self.assertTrue(isinstance(res_subgroup_feature_non.result, CombinedTestStatistics))
        self.assertTrue(isinstance(res_subgroup_feature_has.result, CombinedTestStatistics))
        self.assertLess(res_subgroup_feature_non.result.corrected_test_statistics.statistical_power,
                        res_subgroup_feature_non.result.original_test_statistics.statistical_power)
        self.assertLess(res_subgroup_feature_has.result.corrected_test_statistics.statistical_power,
                        res_subgroup_feature_has.result.original_test_statistics.statistical_power)
        self.assertIsNone(res_subgroup_feature_one.result.original_test_statistics)
        self.assertIsNone(res_subgroup_feature_one.result.corrected_test_statistics)


class OutlierFilteringTestCases(ExperimentTestCase):
    """ Test outlier filtering and quantile filtering. """

    def test_quantile_filtering_multiple_columns(self):
        exp = self.getExperiment()
        flags = exp._quantile_filtering(
            self.data,
            kpis=[
                'normal_same',
                'normal_shifted',
                'normal_shifted_by_feature',
                'normal_unequal_variance'
            ],
            percentile=99.0,
            threshold_type='upper'
        )
        self.assertEqual(len(flags[flags==True]), 386)

    def test_outlier_filtering_lower_threshold(self):
        exp = self.getExperiment()
        data = exp.outlier_filter(
            self.data,
            kpis=[
                'normal_same',
                'normal_shifted',
                'normal_shifted_by_feature',
                'normal_unequal_variance'
            ],
            percentile=0.1,
            threshold_type='lower'
        )
        self.assertEqual(len(self.data) - len(data), exp.metadata['filtered_entities_number'])

    def test_outlier_filtering_unsupported_kpi(self):
        exp = self.getExperiment()
        with self.assertRaises(KeyError):
            exp.outlier_filter(self.data, kpis=['revenue'])

    def test_outlier_filtering_unsupported_percentile(self):
        exp = self.getExperiment()
        with self.assertRaises(ValueError):
            exp.outlier_filter(self.data, kpis=['normal_same'], percentile=101.0)

    def test_outlier_filtering_unsupported_threshold_kind(self):
        exp = self.getExperiment()
        with self.assertRaises(ValueError):
            exp.outlier_filter(self.data, kpis=['normal_same'], threshold_type='uppper')

    def test_outlier_filtering_high_filtering_percentage(self):
        exp = self.getExperiment()
        with warnings.catch_warnings(record=True) as w:
            exp.outlier_filter(self.data, kpis=['normal_same'], percentile=97.9)
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, UserWarning))


class HelperMethodsTestCases(ExperimentTestCase):
    """ Test other helper methods. """

    # Test is_valid_for_analysis
    def test_is_valid_for_analysis(self):
        exp = self.getExperiment()
        is_valid = exp._is_valid_for_analysis(self.data, self.test_normal_same)
        self.assertTrue(is_valid)

        is_not_valid = exp._is_valid_for_analysis(self.data, self.test_nonsense_variant)
        self.assertFalse(is_not_valid)

    # Test _get_weights
    def test_get_weights(self):
        exp = self.getExperiment()
        res = exp._get_weights(self.data, self.test_normal_same, 'B')
        self.assertEqual(res, 1.0)

    # Test _get_weights for derived kpis
    def test_get_weights_derived_kpi(self):
        exp = self.getExperiment()
        self.derived_kpi.make_derived_kpi(self.data)
        res = exp._get_weights(self.data, self.test_derived_kpi, 'B')
        self.assertTrue(isinstance(res, pd.Series))

    # Test re-weighting trick with hardcoded data
    def test_get_weights_hardcoded_data(self):
        ndecimals = 5
        exp = Experiment(self.metadata)
        self.derived_kpi.make_derived_kpi(self.data_dummy_df)
        res = exp._get_weights(self.data_dummy_df, self.test_derived_kpi, 'B')
        self.assertAlmostEqual(res.iloc[0], 1.33333, ndecimals)
        self.assertAlmostEqual(res.iloc[1], 0.66667, ndecimals)

    # Test re-weighting trick with hardcoded data with NaN values
    def test_get_weights_hardcoded_data_with_nan(self):
        ndecimals = 5
        exp = Experiment(self.metadata)
        self.derived_kpi.make_derived_kpi(self.data_dummy_df_with_nan)
        res = exp._get_weights(self.data_dummy_df_with_nan, self.test_derived_kpi, 'B')
        self.assertAlmostEqual(res.iloc[0], 1.33333, ndecimals)
        self.assertAlmostEqual(res.iloc[1], 0.66667, ndecimals)


if __name__ == '__main__':
    unittest.main()



"""Main ExpAn module that contains core and data modules."""

from __future__ import absolute_import

import logging.config

from expan.core import *
from expan.core.version import __version__
from expan.data import *

__all__ = ["core", "data"]
logging.basicConfig(level=logging.DEBUG)



import unittest
import warnings

import numpy as np
from scipy import stats

import expan.core.statistics as statx
from expan.core.util import find_value_by_key_with_condition
from .util import *


class StatisticsTestCase(unittest.TestCase):
    def setUp(self):
        np.random.seed(0)
        data_dir = os.path.dirname(os.path.realpath(__file__)) + '/data'
        self.samples = get_norm_temp_data(data_dir)
        self.metrics = get_framingham_data(data_dir)
        self.rand_s1 = np.random.normal(loc=0, size=1000)
        self.rand_s2 = np.random.normal(loc=0.1, size=1000)

    def tearDown(self):
        pass


class DeltaTestCases(StatisticsTestCase):
    def test__delta__not_providing_data_fails(self):
        """ Value error raised when not providing data. """
        # Check if error is raised for None data
        with self.assertRaises(ValueError):
            statx.delta(self.samples.temperature, None)
        # Check if error is raised for None data
        with self.assertRaises(ValueError):
            statx.delta(None, self.samples.temperature)

    def test__delta__computation_assumed_normal(self):
        """ Result of delta() assuming normality equals expected result. """
        # Computing delta assumed normal
        res = statx.delta(
            self.samples.temperature[self.samples.gender == 1],
            self.samples.temperature[self.samples.gender == 2],
            assume_normal=True)
        # Checking if mean has right value
        self.assertAlmostEqual(res.delta, -0.28923076923075541)

        value025 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 2.5, 'value')
        value975 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 97.5, 'value')

        # Checking if lower percentile has right value
        self.assertAlmostEqual(value025, -0.53770569567692295)
        # Checking if uper percentile has right value
        self.assertAlmostEqual(value975, -0.040755842784587965)
        # Checking if sample size 1 is correct
        self.assertEqual(res.treatment_statistics.sample_size, 65)
        # Checking if sample size 2 is correct
        self.assertEqual(res.control_statistics.sample_size, 65)

    def test__delta__2percentiles_no_tests(self):
        """ Percentiles of delta() for sga are corrected for no tests (1 as a default). """
        res = statx.delta(
            self.samples.temperature[self.samples.gender == 1],
            self.samples.temperature[self.samples.gender == 2])

        value025 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 2.5, 'value')
        value975 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 97.5, 'value')

        self.assertAlmostEqual(value025, -0.53770569567692295)
        self.assertAlmostEqual(value975, -0.040755842784587965)

    def test__delta__nan_handling(self):
        """ Test correct handling of nans. (ignored). """
        res = statx.delta(self.rand_s1, self.rand_s2)
        self.assertEqual(res.treatment_statistics.sample_size, 1000)
        self.assertEqual(res.control_statistics.sample_size, 1000)

        r1 = self.rand_s1.copy()
        r1[90:] = np.nan
        res = statx.delta(r1, self.rand_s2)

        self.assertAlmostEqual (res.delta, -0.1, 1)
        self.assertEqual       (res.treatment_statistics.sample_size, 90)
        self.assertEqual       (res.control_statistics.sample_size, 1000)

    def test__delta__computation_not_assumed_normal(self):
        """ Result of delta() not assuming normality equals expected result. """
        # Computing delta not assumed normal
        res = statx.delta(
            self.samples.temperature[self.samples.gender == 1],
            self.samples.temperature[self.samples.gender == 2],
            assume_normal=True)
        # Checking if mean has right value
        self.assertAlmostEqual(res.delta, -0.28923076923075541)

        value025 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 2.5, 'value')
        value975 = find_value_by_key_with_condition(res.confidence_interval, 'percentile', 97.5, 'value')

        # Checking if lower percentile has right value
        self.assertAlmostEqual(value025, -0.53770569567692295)
        # Checking if uper percentile has right value
        self.assertAlmostEqual(value975, -0.040755842784587965)
        # Checking if sample size 1 is correct
        self.assertEqual(res.treatment_statistics.sample_size, 65)
        # Checking if sample size 2 is correct
        self.assertEqual(res.control_statistics.sample_size, 65)


class SampleSizeTestCases(StatisticsTestCase):
    def test__sample_size__empty_list_numeric(self):
        """ Empty list returns 0. """
        self.assertEqual(statx.sample_size([]), 0)

    def test__sample_size__list_numeric(self):
        """ Result of sample_size() is number of elements of a list. """
        x = [1, 1, 2, 5, 8]
        self.assertEqual(statx.sample_size(x), 5)

    def test__sample_size__nparray_numeric_with_nan(self):
        """ Result of sample_size() is number of elements of numpy array minus number of NaNs. """
        x = np.array([1, 1, np.nan, 2, np.nan, 5, 8])
        self.assertEqual(statx.sample_size(x), 5)

    def test__sample_size__list_categorical(self):
        """ Result of sample_size() is number of elements of a list of categorical data. """
        x = ['1', '1', '3', '2', '6', '5', '8']
        self.assertEqual(statx.sample_size(x), 7)

    def test__sample_size__pdseries_categorical(self):
        """Result of sample_size() is number of elements of a pandas series with categorical data. """
        x = pd.Series(['1', '7', '2', '5', '8', '0'])
        self.assertEqual(statx.sample_size(x), 6)

    def test__sample_size__pdseries_categorical_with_na(self):
        """ Result of sample_size() is number of elements of a pandas series with categorical data including NAs. """
        x = ['1', '1', 'NA', '2', 'NA', '5', '8']
        self.assertEqual(statx.sample_size(x), 5)


class EstimateSampleSizeTestCases(StatisticsTestCase):
    def test__estimate_sample_size__dataframe(self):
        """ Result of estimate_sample_size() is a series of estimated sample sizes. """
        x = pd.DataFrame({'sample_1': [1, 7, 8, 9, 3, 4, 2, 0], 'sample_2': [5, 4, 3, 2, 2, 2, 1, 0]})
        res = statx.estimate_sample_size(x=x, mde=0.01, r=1.0)
        self.assertEqual(int(res['sample_1']), 197405)
        self.assertEqual(int(res['sample_2']), 142130)

    def test__estimate_sample_size__series(self):
        """ Result of estimate_sample_size() is estimated sample size. """
        x = pd.Series([1, 7, 8, 9, 3, 4, 2, 0])
        self.assertEqual(int(statx.estimate_sample_size(x=x, mde=0.01, r=1.0)), 197405)

    def test__estimate_sample_size__x_type_error(self):
        """ Method estimate_sample_size raises TypeError since x is a list. """
        x = [1, 7, 8, 9, 3, 4, 2, 0]
        self.assertRaises(TypeError, statx.estimate_sample_size, x=x, mde=0.01, r=1.0)

    def test__estimate_sample_size__r_value_error(self):
        """ Method estimate_sample_size raises ValueError since r is 0. """
        x = pd.Series([1, 7, 8, 9, 3, 4, 2, 0])
        self.assertRaises(ValueError, statx.estimate_sample_size, x=x, mde=0.01, r=0.0)


class BootstrapTestCases(StatisticsTestCase):
    def test__bootstrap__computation(self):
        """ Result of bootstrap() equals expected result. """
        # Computing bootstrap
        result1 = statx.bootstrap(self.samples.temperature, [0], min_observations=1)
        # Checking if lower percentile of result2 is correct
        self.assertAlmostEqual(result1[0][2.5], 98.1220, 2)
        # Checking if upper percentile of result2 is correct
        self.assertAlmostEqual(result1[0][97.5], 98.3708, 2)
        # Checking if no bootstrap data was passed
        self.assertIsNone(result1[1])

        # Defining data and computing bootstrap
        zero_3 = np.array([0., 0., 0.])
        one_3 = np.array([1., 1., 1.])
        result2 = statx.bootstrap(zero_3, one_3, min_observations=3)
        # Checking if lower percentile of result2 is correct
        self.assertEqual(result2[0][2.5], -1.0)
        # Checking if upper percentile of result2 is correct
        self.assertEqual(result2[0][97.5], -1.0)
        # Checking if no bootstrap data was passed
        self.assertIsNone(result2[1])

        # Defining data and computing bootstrap
        sample1 = self.samples.temperature[self.samples.gender == 1]
        sample2 = self.samples.temperature[self.samples.gender == 2]
        result3 = statx.bootstrap(sample1, sample2)
        # Checking if lower percentile of result3 is correct
        self.assertAlmostEqual(result3[0][2.5], -0.53384615384615619)
        # Checking if upper percentile of result3 is correct
        self.assertAlmostEqual(result3[0][97.5], -0.049192307692299965)
        # Checking if no bootstrap data was passed
        self.assertIsNone(result3[1])


class PooledStdTestCases(StatisticsTestCase):
    def test__pooled_std__variances_differ_too_much_error(self):
        """ Warning raised when variances differ too much. """
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            statx.pooled_std(0.25, 4, 0.5, 4)
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, UserWarning))
            self.assertTrue('variances differ' in str(w[-1].message))

    def test__pooled_std__computation(self):
        """ Result of pooled_std() equals expected result. """
        # Define subset of data for test
        sbp = self.metrics.loc['systolic_bp', :]
        # Computation of pooled standard deviation
        result1 = statx.pooled_std(sbp.men.s, sbp.men.n, sbp.women.s, sbp.women.n)
        # Checking if result1 is correct
        self.assertAlmostEqual(result1, 19, 1)

        # Computation of pooled standard deviation
        result2 = statx.pooled_std(9.7, 6, 12.0, 4)
        # Checking pooled_std() on small sample subset
        self.assertAlmostEqual(result2, 10.6, 1)


class NormalDifferenceTestCases(StatisticsTestCase):
    def test__normal_difference__computation(self):
        """ Result of normal_difference() equals expected result. """
        # Define subset of data for first test
        sbp = self.metrics.loc['systolic_bp', :]
        # computation of normal difference
        result1 = statx.normal_difference(sbp.men.m, sbp.men.s, sbp.men.n, sbp.women.m, sbp.women.s, sbp.women.n)
        # Checking if lower percentile of result1 is correct
        self.assertAlmostEqual(result1[2.5], 0.44582598543756413)
        # Checking if upper percentile of result1 is correct
        self.assertAlmostEqual(result1[97.5], 2.9541740145624127)

        # Define subset of data for second test
        clst = self.metrics.loc['serum_cholesterol', :]
        # Computation of normal difference
        result2 = statx.normal_difference(clst.men.m, clst.men.s, clst.men.n,
                                          clst.women.m, clst.women.s, clst.women.n)
        # Checking if lower percentile of result2 is correct
        self.assertAlmostEqual(result2[2.5], -17.159814380797162)
        # Checking if upper percentile of result2 is correct
        self.assertAlmostEqual(result2[97.5], -12.240185619202816)

        # test subsample of systolic blood pressure. Example from:
        # http://sphweb.bumc.bu.edu/otlt/MPH-Modules/BS/BS704_Confidence_Intervals/BS704_Confidence_Intervals5.html
        # Computation of normal difference
        result3 = statx.normal_difference(117.5, 9.7, 6, 126.8, 12., 4)
        # Checking if lower percentile of result3 is correct
        self.assertAlmostEqual(result3[2.5], -25.10960582643531)
        # Checking if upper percentile of result3 is correct
        self.assertAlmostEqual(result3[97.5], 6.5096058264353118)


class NormalSampleDifferenceTestCases(StatisticsTestCase):
    def test__normal_sample_difference__computation(self):
        # Defining data and computing normal difference
        sample1 = self.samples.temperature[self.samples.gender == 1]
        sample2 = self.samples.temperature[self.samples.gender == 2]
        # Computation of normal sample difference
        result1 = statx.normal_sample_difference(sample1, sample2)
        # Checking if lower percentile of result1 is correct
        self.assertAlmostEqual(result1[2.5], -0.53770569567692295)
        # Checking if upper percentile of result1 is correct
        self.assertAlmostEqual(result1[97.5], -0.040755842784587965)


class StatisticalPowerTestCases(StatisticsTestCase):
    def test_compute_statistical_power(self):
        float_precision = 2
        # Confirm with pre-computed value by hand via power analysis
        alpha = 0.05
        beta  = 0.2
        sigma = 1
        mean1 = 1
        mean2 = 0
        n1    = 13
        n2    = 12
        z_1_minus_alpha = stats.norm.ppf(1-alpha)

        power = statx.compute_statistical_power(mean1, sigma, n1, mean2, sigma, n2, z_1_minus_alpha)
        self.assertAlmostEqual(power, 1-beta, float_precision)
        
    def test_zero_pooled_std(self):
        with self.assertRaises(ValueError):
            statx.compute_statistical_power_from_samples([1, 1, 1, 1],
                    [2, 2, 2, 2])


class PValueTestCases(StatisticsTestCase):
    def test_compute_p_value(self):
        float_precision = 2
        # Confirm with pre-computed value by hand via looking up t table
        # Given the following values, pooled std and standard error should both be 1.
        # By looking up from two-tailed t-table of degree of freedom 2, t=4.303 will lead to p=0.05.
        sigma      = 1
        mean1      = 4.303
        mean2      = 0
        n1         = 2
        n2         = 2
        expected_p = 0.05

        p = statx.compute_p_value(mean1, sigma, n1, mean2, sigma, n2)
        self.assertAlmostEqual(p, expected_p, float_precision)


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
import expan.core.util as util


class UtilFunctionsTestCase(unittest.TestCase):
    def test_find_value_by_key_with_condition(self):
        list_of_dicts = [{'bla': 1, 'blu': 2},
                         {'bla': 3, 'blu': 4},
                         {'bla': 5, 'blu': 6}]
        self.assertEqual(util.find_value_by_key_with_condition(list_of_dicts, 'bla', 5, 'blu'), 6)

        with self.assertRaises(IndexError):
            util.find_value_by_key_with_condition(list_of_dicts, 'bla', 'value_not_exist', 'blu')
        with self.assertRaises(KeyError):
            util.find_value_by_key_with_condition(list_of_dicts, 'bla', 5, 'key_not_exist')

    def test_is_nan(self):
        a = float('nan')   # NaN
        self.assertTrue(util.is_nan(a))
        b = None
        self.assertFalse(util.is_nan(b))
        c = 2
        self.assertFalse(util.is_nan(c))
        d = 'str'
        self.assertFalse(util.is_nan(d))

    def test_drop_nan(self):
        nan = float('nan')   # NaN

        array_with_nan_1d = np.array([nan, nan, 1, 2, 3])
        returned_array_1d = util.drop_nan(array_with_nan_1d)
        self.assertTrue(all(returned_array_1d == np.array([1, 2, 3])))

        array_with_nan_2d = np.array([[nan], [nan], [1], [2], [3]])
        returned_array_2d = util.drop_nan(array_with_nan_2d)
        print(returned_array_2d)
        self.assertTrue(all(returned_array_2d == np.array([[1], [2], [3]])))

        array_without_nan_1d = np.array([1, 2])
        returned_array = util.drop_nan(array_without_nan_1d)
        self.assertTrue(all(returned_array == np.array([1, 2])))


if __name__ == '__main__':
    unittest.main()



import os
import pandas as pd
import numpy as np


def get_norm_temp_data(fdir, fname='normtemp.dat.txt.gz'):
    """ Gets normtemp dataset: normtemp.dat.txt.gz.
    Data retrieved on 2015/02/18 from: http://www.amstat.org/publications/jse/jse_data_archive.htm

    :param fdir: directory containing the data
    :type  fdir: str
    :param fname: data file name
    :type  fname: str
    
    :return data frame containing test normtemp data
    :rtype  pd.DataFrame
    """

    # Read data from csv to pd.dataFrame
    data = pd.read_csv(
        os.path.join(fdir, fname),
        delim_whitespace=True,
        header=None,
        skip_blank_lines=True,
        names=['temperature', 'gender', 'heartrate'],
    )
    return data


def get_framingham_data(fdir, fname='framingham_heart_study_exam7.csv'):
    """ Gets framingham heart study dataset: framingham_heart_study_exam7.csv:
    Data retrieved on 2015/10/28 from:
    http://sphweb.bumc.bu.edu/otlt/MPH-Modules/BS/BS704_Confidence_Intervals/BS704_Confidence_Intervals5.html

    :param fdir: directory containing the data
    :type  fdir: str
    :param fname: data file name
    :type  fname: str
    
    :return data frame containing test framingham heart study dataset
    :rtype  pd.DataFrame
    """
    data = pd.read_csv(os.path.join(fdir, fname),
                       header=[0, 1],
                       index_col=0)
    data.index.name = 'metric'
    return data


def get_test_data_revenue_order():
    """ Small test data for testing with kpis revenue and number of orders.
    
    :return data frame containing test dataset
    :rtype  pd.DataFrame
    """
    test_data = np.array([['fullVisitorId', 'visitorId', 'date', 'entity', 'variant',
                           'revenue', 'orders', 'appdomain', 'device_type', 'gender'],
                           ['fv1', 'v1', '2017-10-01', 'client-1', 'variant1', 2.3, 4.5, 'AT', 'mobile', 'female'],
                           ['fv1', 'v2', '2017-10-01', 'client-1', 'variant1', 1.2, 0.5, 'AT', 'desktop', 'female'],
                           ['fv1', 'v1', '2017-10-01', 'client-2', 'variant1', 2.1, 3.5, 'AT', 'mobile', 'male'],
                           ['fv1', 'v2', '2017-10-01', 'client-2', 'variant1', 1.1, 0.7, 'AT', 'desktop', 'male'],
                           ['fv1', 'v1', '2017-10-01', 'client-3', 'variant2', 2.3, 4.5, 'AT', 'mobile', 'male'],
                           ['fv1', 'v2', '2017-10-01', 'client-3', 'variant2', 1.2, 0.5, 'AT', 'desktop', 'male'],
                           ['fv1', 'v1', '2017-10-01', 'client-4', 'variant2', 2.1, 3.5, 'AT', 'mobile', 'female'],
                           ['fv1', 'v2', '2017-10-01', 'client-4', 'variant2', 1.1, 0.7, 'AT', 'desktop',
                            'female']])
    return pd.DataFrame(data=test_data[1:, 1:], columns=test_data[0, 1:]).convert_objects(convert_numeric=True)



import subprocess

__version__ = "1.1.0"


def version_numbers():
    """ Returns ExpAn version. """
    return [int(x) for x in __version__.split('.')]


def git_commit_count():
    """ Returns the output of `git rev-list --count HEAD` as an int.
    Note: http://programmers.stackexchange.com/a/151558 """

    return int(subprocess.check_output(["git", "rev-list", "--count", "HEAD"]))


def git_latest_commit():
    """ Returns output of `git rev-parse HEAD`.
    Note: http://programmers.stackexchange.com/a/151558. """

    return subprocess.check_output(["git", "rev-parse", "HEAD"]).strip()


def version(format_str='{short}'):
    """ Returns current version number in specified format.
    
    :param format_str: format string for the version
    :type  format_str: str

    :return: version number in the specified format
    :rtype: str
    """
    major, minor, patch = version_numbers()

    format_dict = {
        'major': major,
        'minor': minor,
        'patch': patch,
        'short': 'v{:d}.{:d}.{:d}'.format(major, minor, patch),
    }

    format_str = format_str.replace(
        '{commits', '{commit_count').replace(
        '{num_commits', '{commit_count').replace(
        '{last_commit', '{latest_commit').replace(
        '{commit', '{latest_commit').replace(
        '{hash', '{latest_commit').replace(
        '{HEAD', '{latest_commit').replace(
        '{long}', 'v{major:d}.{minor:d}.{patch:d}.{commit_count:d}-r{latest_commit}')

    if ('{commit_count' in format_str) or ('{latest_commit' in format_str):
        format_dict['commit_count'] = git_commit_count()
        format_dict['latest_commit'] = git_latest_commit()

    return format_str.format(**format_dict)



"""ExpAn core module.
"""

from __future__ import absolute_import

from expan.core.version import __version__, version

__all__ = ["binning", "early_stopping", "experiment", "statistics", "util",
           "version", "results", "correction", "statistical_test"]

print(('ExpAn core init: {}'.format(version())))



#!/usr/bin/env python

import re
from setuptools import setup, find_packages

def parse_requirements(filename):
    """ load requirements from a pip requirements file """
    lineiter = (line.strip() for line in open(filename))
    return [line for line in lineiter if line and not line.startswith("#")]

try:
    requirements = parse_requirements('requirements.txt')
except OSError:
	requirements = []

with open('README.rst') as readme_file:
	readme = readme_file.read()

with open('CHANGELOG.rst') as history_file:
	history = history_file.read()

with open('expan/core/version.py', 'r') as fd:
	version = re.search(
		r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
		fd.read(),
		re.MULTILINE).group(1)

if not version:
	raise RuntimeError('Cannot find version information')

test_requirements = [
    # the latest pytest 3.1.0 will lead to failure of some unit tests in python 2.7
	'pytest==3.0.7'
]

setup(
	name='expan',
	version=version,
	description="Experiment Analysis Library",
	long_description=readme + '\n\n' + history,
	author="Zalando SE",
	author_email='octopus@zalando.de',
	url='https://github.com/zalando/expan',
	packages=find_packages(),
	package_dir={'expan': 'expan'},
	include_package_data=True,
	install_requires=requirements,
	license="MIT",
	zip_safe=False,
	keywords='expan',
	entry_points={
		'console_scripts': [
			'expan = expan.cli.cli:main'
		]
	},
	classifiers=[
		'Development Status :: 3 - Alpha',
		'Intended Audience :: Developers',
		'Intended Audience :: Science/Research',
		'License :: OSI Approved :: MIT License',
		'Natural Language :: English',
		"Programming Language :: Python :: 2",
		# 'Programming Language :: Python :: 2.6',
		'Programming Language :: Python :: 2.7',
		'Programming Language :: Python :: 3',
		# 'Programming Language :: Python :: 3.3',
		'Programming Language :: Python :: 3.4',
		'Programming Language :: Python :: 3.5',
	],
	test_suite='tests',
	tests_require=test_requirements
)



import numpy as np


def benjamini_hochberg(false_discovery_rate, original_p_values):
    """ Benjamini-Hochberg procedure.
    
    :param false_discovery_rate: proportion of significant results that are actually false positives
    :type  false_discovery_rate: float
    :param original_p_values: p values from all the tests
    :type  original_p_values: list[float]
    
    :return: new critical value (i.e. the corrected alpha)
    :rtype: float
    """
    p_values_sorted = np.sort(np.asarray(original_p_values))
    number_tests = len(original_p_values)
    significant_ranks = [i for i, val in enumerate(p_values_sorted, 1) if val <= i * false_discovery_rate / number_tests]
    rank = np.max(significant_ranks) if significant_ranks else 1
    return rank * false_discovery_rate / number_tests


def bonferroni(false_positive_rate, original_p_values):
    """ Bonferrnoi correction.
    
    :param false_positive_rate: alpha value before correction
    :type  false_positive_rate: float
    :param original_p_values: p values from all the tests
    :type  original_p_values: list[float]
    
    :return: new critical value (i.e. the corrected alpha)
    :rtype: float
    """
    return false_positive_rate / len(original_p_values)



import json
import logging
from enum import Enum

import numpy as np
import pandas as pd
from datetime import datetime, timedelta

logger = logging.getLogger(__name__)


class JsonSerializable(object):
    """ Interface for serializable classes."""
    def toJson(self):
        return json.dumps(self, default=lambda o: o.name if isinstance(o, Enum) else o.__dict__, sort_keys=True, indent=4)

    def __repr__(self):
        return self.toJson()


def find_value_by_key_with_condition(items, condition_key, condition_value, lookup_key):
    """ Find the value of lookup key where the dictionary contains condition key = condition value.
    
    :param items: list of dictionaries
    :type  items: list
    :param condition_key: condition key
    :type  condition_key: str
    :param condition_value: a value for the condition key
    :param lookup_key: lookup key or key you want to find the value for
    :type  lookup_key: str
    
    :return: lookup value or found value for the lookup key
    """
    return [item[lookup_key] for item in items if item[condition_key] == condition_value][0]


def is_nan(obj):
    """ Checks whether the input is NaN. It uses the trick that NaN is not equal to NaN."""
    return obj != obj


def drop_nan(array):
    """ Drop Nan values from the given numpy array.
    
    :param array: input array
    :type  array: np.ndarray
    
    :return: a new array without NaN values
    :rtype: np.ndarray
    """
    if array.ndim == 1:
        return array[~np.isnan(array)]
    elif array.ndim == 2:
        return array[~np.isnan(array).any(axis=1)]


def generate_random_data():
    """ Generate random data for two variants. It can be used in unit tests or demo. """
    np.random.seed(42)
    size = 10000

    data = pd.DataFrame()
    data['entity'] = list(range(size))
    data['variant'] = np.random.choice(['A', 'B'], size=size, p=[0.6, 0.4])

    data['normal_same'] = np.random.normal(size=size)
    data['normal_shifted'] = np.random.normal(size=size)

    size_shifted_B = data['normal_shifted'][data['variant'] == 'B'].shape[0]
    data.loc[data['variant'] == 'B', 'normal_shifted'] = np.random.normal(loc=1.0, size=size_shifted_B)

    data['feature'] = np.random.choice(['has', 'non'], size=size)
    data.loc[0, 'feature'] = 'feature that only has one data point'
    data['normal_shifted_by_feature'] = np.random.normal(size=size)

    ii = (data['variant'] == 'B') & (data['feature'] == 'has')
    randdata_shifted_mean = np.random.normal(loc=1.0, size=sum(ii == True))
    data.loc[ii, 'normal_shifted_by_feature'] = randdata_shifted_mean

    data['treatment_start_time'] = np.random.choice(list(range(10)), size=size)
    data['normal_unequal_variance'] = np.random.normal(size=size)

    size_unequalvar_B = data['normal_unequal_variance'][data['variant'] == 'B'].shape[0]
    data.loc[data['variant'] == 'B', 'normal_unequal_variance'] = np.random.normal(scale=10, size=size_unequalvar_B)

    # Add date column
    d1 = datetime.strptime('2015-01-01', '%Y-%m-%d')
    d2 = datetime.strptime('2016-03-01', '%Y-%m-%d')
    date_col = []
    delta = d2 - d1
    for i in range(delta.days * 24 + 1):
        date_col.append((d1 + timedelta(hours=i)).strftime('%Y-%m-%d'))
    data['date'] = date_col[:size]

    metadata = {
        'primary_KPI': 'normal_shifted',
        'source': 'simulated',
        'experiment': 'random_data_generation'
    }
    return data, metadata



import unittest
import numpy as np

from expan.core.statistical_test import *
from expan.core.util import generate_random_data


class StatisticalTestCase(unittest.TestCase):
    def setUp(self):
        np.random.seed(41)
        self.data, self.metadata = generate_random_data()

        # simple statistical test
        self.test_kpi = KPI('normal_same')
        self.test_variants = Variants('variant', 'A', 'B')
        self.test_normal_same = StatisticalTest(self.data, self.test_kpi, [], self.test_variants)

    def tearDown(self):
        pass

    def test_setup_single_test(self):
        variants = Variants('variant', 'control', 'treatment')
        mobile = FeatureFilter('device_type', 'mobile')
        kpi = KPI('revenue')

        test_revenue_overall = StatisticalTest(self.data, kpi, [], variants)
        test_revenue_mobile = StatisticalTest(self.data, kpi, [mobile], variants)

        self.assertEqual(test_revenue_overall.kpi.name, "revenue")
        self.assertEqual(test_revenue_overall.variants.control_name, "control")
        self.assertEqual(test_revenue_overall.variants.treatment_name, "treatment")

        self.assertEqual(test_revenue_mobile.kpi.name, "revenue")
        self.assertEqual(test_revenue_mobile.variants.control_name, "control")
        self.assertEqual(test_revenue_mobile.variants.treatment_name, "treatment")
        self.assertEqual(test_revenue_mobile.features[0].column_name, "device_type")
        self.assertEqual(test_revenue_mobile.features[0].column_value, "mobile")

    def test_setup_multiple_test_suite(self):
        variants = Variants('variant', 'control', 'treatment')
        mobile = FeatureFilter('device_type', 'mobile')
        desktop = FeatureFilter('device_type', 'desktop')
        tablet = FeatureFilter('device_type', 'tablet')
        kpi = KPI('revenue')

        test_revenue_overall = StatisticalTest(self.data, kpi, [], variants)
        test_revenue_mobile = StatisticalTest(self.data, kpi, [mobile], variants)
        test_revenue_desktop = StatisticalTest(self.data, kpi, [desktop], variants)
        test_revenue_tablet = StatisticalTest(self.data, kpi, [tablet], variants)

        tests = [test_revenue_overall, test_revenue_mobile, test_revenue_desktop, test_revenue_tablet]
        multi_test_suite = StatisticalTestSuite(tests, CorrectionMethod.BH)

        self.assertEqual(multi_test_suite.size, 4)
        self.assertEqual(multi_test_suite.correction_method, CorrectionMethod.BH)

    def test_make_derived_kpi(self):
        numerator = "normal_same"
        denominator = "normal_shifted"
        derived_kpi_name = "derived_kpi_one"
        DerivedKPI(derived_kpi_name, numerator, denominator).make_derived_kpi(self.data)

        # checks if column with the derived kpi was created
        self.assertTrue(derived_kpi_name in self.data.columns)

        # checks if values of new columns are of type float
        self.assertTrue(all(isinstance(kpi_value, float) for kpi_value in self.data[derived_kpi_name]))

    def test_get_variant(self):
        control = self.test_normal_same.variants.get_variant(
            self.data, self.test_normal_same.variants.control_name)[self.test_normal_same.kpi.name]
        self.assertEqual(len(control), 6108)
        self.assertTrue(isinstance(control, pd.Series))



import unittest
from os import rmdir, makedirs, getcwd, remove, walk
from os.path import dirname, join, realpath, exists

import simplejson as json

import expan.core.util
import expan.data.csv_fetcher as csv_fetcher

__location__ = realpath(join(getcwd(), dirname(__file__)))

TEST_FOLDER = join(__location__, 'test_folder')


class CsvFetcherTestCase(unittest.TestCase):
    def setUp(self):
        # create test folder
        if not exists(TEST_FOLDER):
            makedirs(TEST_FOLDER)
        # generate data and metadata
        (data, metadata) = expan.core.util.generate_random_data()
        # save data to .csv.gz file in test folder
        data.to_csv(path_or_buf=join(TEST_FOLDER, 'data.csv.gz'), compression='gzip')
        # save metadata to .json file in test folder
        with open(join(TEST_FOLDER, 'metadata.json'), 'w') as f:
            json.dump(metadata, f)

    def tearDown(self):
        # remove all test files and test folder
        for root, dirs, files in walk(TEST_FOLDER, topdown=False):
            for name in files:
                remove(join(root, name))
            for name in dirs:
                rmdir(join(root, name))
        rmdir(TEST_FOLDER)

    def test_csv_fetcher(self):
        csv_fetcher.get_data(TEST_FOLDER)
        with self.assertRaises(AssertionError):
            csv_fetcher.get_data(join(__location__, '..'))



import unittest
import numpy as np

from expan.core.results import *
from expan.core.statistical_test import *
from expan.core.util import generate_random_data


class StatisticalTestCase(unittest.TestCase):
    def setUp(self):
        np.random.seed(41)
        self.data, self.metadata = generate_random_data()

        np.random.seed(41)
        self.control_statistics   = SampleStatistics(1000, 0.0, 1.0)
        self.treatment_statistics = SampleStatistics(1200, 1.0, 1.0)
        self.delta                = 1.0
        self.p                    = 0.04
        self.statistical_power    = 0.8
        self.confidence_interval  = {2.5: 0.1, 97.5: 1.1}
        self.stop                 = True

        self.corrected_p          = 0.02
        self.corrected_ci         = {1.0: -0.7, 99.0: 0.7}
        self.corrected_stop       = False

        self.simple_stats = SimpleTestStatistics(self.control_statistics,
                                                 self.treatment_statistics,
                                                 self.delta,
                                                 self.confidence_interval,
                                                 self.p,
                                                 self.statistical_power)

        self.simple_stats_corrected = SimpleTestStatistics(self.control_statistics,
                                                           self.treatment_statistics,
                                                           self.delta,
                                                           self.corrected_ci,
                                                           self.corrected_p,
                                                           self.statistical_power)

        self.es_stats = EarlyStoppingTestStatistics(self.control_statistics,
                                                    self.treatment_statistics,
                                                    self.delta,
                                                    self.confidence_interval,
                                                    self.p,
                                                    self.statistical_power,
                                                    self.stop)

        self.es_stats_corrected = EarlyStoppingTestStatistics(self.control_statistics,
                                                              self.treatment_statistics,
                                                              self.delta,
                                                              self.corrected_ci,
                                                              self.corrected_p,
                                                              self.statistical_power,
                                                              self.corrected_stop)

        variants = Variants('variant', 'control', 'treatment')
        self.correction_method = CorrectionMethod.BONFERRONI

        kpi = KPI('revenue')
        self.statistical_test = StatisticalTest(self.data, kpi, [], variants)
        self.statistical_test_result = StatisticalTestResult(self.statistical_test, self.simple_stats_corrected)

        test_result1 = StatisticalTestResult(self.statistical_test,
                                             CombinedTestStatistics(self.simple_stats, self.simple_stats_corrected))
        test_result2 = StatisticalTestResult(self.statistical_test,
                                             CombinedTestStatistics(self.es_stats, self.es_stats_corrected))
        test_results = [test_result1, test_result2]
        self.statistical_test_results = MultipleTestSuiteResult(test_results, self.correction_method)

    def tearDown(self):
        pass

    def test_simple_test_statistics(self):
        self.assertAlmostEqual(self.simple_stats.delta, 1.0)
        self.assertEqual(len(self.simple_stats.confidence_interval), 2)

    def test_early_stopping_test_statistics(self):
        self.assertAlmostEqual(self.es_stats.delta, 1.0)
        self.assertEqual(len(self.es_stats.confidence_interval), 2)
        self.assertTrue(self.es_stats.stop)

    def test_corrected_test_statistics_simple(self):
        corrected_test_statistics = CombinedTestStatistics(self.simple_stats, self.simple_stats_corrected)
        self.assertEqual(corrected_test_statistics.original_test_statistics.p, 0.04)
        self.assertEqual(corrected_test_statistics.corrected_test_statistics.p, 0.02)

    def test_corrected_test_statistics_early_stopping(self):
        corrected_test_statistics = CombinedTestStatistics(self.es_stats, self.es_stats_corrected)
        self.assertEqual(corrected_test_statistics.original_test_statistics.p, 0.04)
        self.assertEqual(corrected_test_statistics.corrected_test_statistics.p, 0.02)
        self.assertTrue(corrected_test_statistics.original_test_statistics.stop)
        self.assertFalse(corrected_test_statistics.corrected_test_statistics.stop)

    def test_corrected_test_statistics_type_mismatch(self):
        type1 = str(type(self.simple_stats))
        type2 = str(type(self.es_stats_corrected))
        error_msg = "Type mismatch for type " + type1 + " and " + type2
        with self.assertRaisesRegexp(RuntimeError, error_msg):
            CombinedTestStatistics(self.simple_stats, self.es_stats_corrected)

    def test_statistical_test_results(self):
        self.assertEqual(self.statistical_test_result.test.kpi.name, 'revenue')
        self.assertAlmostEqual(self.statistical_test_result.result.delta, 1.0)

    def test_multi_test_suite_results(self):
        self.assertEqual(len(self.statistical_test_results.results), 2)
        self.assertEqual(self.statistical_test_results.correction_method, self.correction_method)



import logging
from os import listdir
from os.path import isfile, join

import pandas as pd
import simplejson as json

from expan.core.experiment import Experiment

logger = logging.getLogger(__name__)


def get_data(folder_path):
    """ Expects as input a folder containing the following files:

    - one .csv or .csv.gz with 'data' in the filename
    - one .json containing 'metadata' in the filename

    Opens the files and uses them to create an Experiment object which it then returns.

    :param folder_path: path to the Experiment data
    :type  folder_path: str
    :return: Experiment object with data
    :rtype:  Experiment
    """
    files = [f for f in listdir(folder_path) if isfile(join(folder_path, f))]

    try:
        assert ('data' in '-'.join(files))
        assert ('metadata' in '-'.join(files))
        data = metadata = None
        for f in files:
            if 'metadata' in f:
                with open(join(folder_path, f), 'r') as input_json:
                    metadata = json.load(input_json)
            elif 'data' in f:
                data = pd.read_csv(join(folder_path, f))
        return data, metadata

    except AssertionError as e:
        logger.error("An error occurred when fetching data from csv file.")
        raise e



""" NB: This module is deprecated. """

import logging
import warnings
from heapq import heapify, heappush, heappop

import numpy as np

from expan.core.util import is_nan

logger = logging.getLogger(__name__)


#------- bin representations -------#
class Bin(object):
    # this is a necessary hack for the buggy implementation of assertItemsEqual in python2
    # see https://stackoverflow.com/a/29690198
    # note that if we only use python3, assertCountEqual in python3 solves this problem
    __hash__ = None

    def __init__(self, bin_type, *repr_args):
        """
        Constructor for a bin object.
        :param id: identifier (e.g. bin number) of the bin
        :param bin_type: "numerical" or "categorical"
        :param repr_args: arguments to represent this bin. 
        args for numerical bin includes lower, upper, lower_closed, upper_closed
        args for categorical bin includes a list of categories for this bin.
        """
        if bin_type == "numerical" and len(repr_args) != 4:
            raise ValueError("args for numerical bin are lower, upper, lower_closed, upper_closed.")
        if bin_type == "categorical" and len(repr_args) != 1 and type(repr_args[0]) is not list:
            raise ValueError("args for categorical bin is a list of categorical values for this bin.")
        self.bin_type = bin_type

        if bin_type == "numerical":
            self.representation = NumericalRepresentation(*repr_args)
        elif bin_type == "categorical":
            self.representation = CategoricalRepresentation(*repr_args)

    def __repr__(self):
        return "\nbin: " + str(self.representation)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __call__(self, data, feature):
        """
        Apply the bin to data.
        :param data: pandas data frame
        :param feature: feature name on which this bin is defined
        :return: subset of input dataframe which belongs to this bin
        """
        return self.representation.apply_to_data(data, feature)


class NumericalRepresentation(object):
    # this is a necessary hack for the buggy implementation of assertItemsEqual in python2
    # see https://stackoverflow.com/a/29690198
    # note that if we only use python3, assertCountEqual in python3 solves this problem
    __hash__ = None

    def __init__(self, lower, upper, lower_closed, upper_closed):
        """
        Constructor for representation of a numerical bin.
        :param upper: upper bound of the bin (exclusive)
        :param lower: lower bound of the bin (inclusive)
        :param lower_closed: boolean indicator whether lower bound is closed
        :param upper_closed: boolean indicator whether upper bound is closed
        """
        self.lower = lower
        self.upper = upper
        self.lower_closed = lower_closed
        self.upper_closed = upper_closed

    def __repr__(self):
        repr = ""
        if self.lower_closed:
            repr += "["
        else:
            repr += "("

        repr += str(self.lower) + ", " + str(self.upper)

        if self.upper_closed:
            repr += "]"
        else:
            repr += ")"
        return repr

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def apply_to_data(self, data, feature):
        """
        Apply the bin to data.
        :param data: pandas data frame
        :param feature: feature name on which this bin is defined
        :return: subset of input dataframe which belongs to this bin
        """
        data_feature_column = data[feature]

        # if either bound is nan, only nans exist in the bin.
        if np.isnan(self.lower) or np.isnan(self.upper):
            return data[np.isnan(data_feature_column)]

        if self.lower_closed:
            filter_lower = (data_feature_column >= self.lower)
        else:
            filter_lower = (data_feature_column > self.lower)
        if self.upper_closed:
            filter_upper = (data_feature_column <= self.upper)
        else:
            filter_upper = (data_feature_column < self.upper)

        return data[filter_lower & filter_upper]


class CategoricalRepresentation(object):
    # this is a necessary hack for the buggy implementation of assertItemsEqual in python2
    # see https://stackoverflow.com/a/29690198
    # note that if we only use python3, assertCountEqual in python3 solves this problem
    __hash__ = None

    def __init__(self, categories):
        """
        Constructor for representation of a categorical bin.
        :param categories: list of categorical values that belong to this bin
        """
        try:
            categories = set(categories)
        except:
            raise ValueError("categorical bin should be represented by an iterable object of categories.")

        self.categories = categories

    def __repr__(self):
        return str(list(self.categories))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def apply_to_data(self, data, feature):
        """
        Apply the bin to data.
        :param data: pandas data frame
        :param feature: feature name on which this bin is defined
        :return: subset of input dataframe which belongs to this bin
        """
        data_feature_column = data[feature]
        return data[data_feature_column.isin(self.categories)]



#------- public methods -------#
def create_bins(data, n_bins):
    """
    Create bins from the data value
    :param data: a list or a 1-dim array of data to determine the bins
    :param n_bins: number of bins to create
    :return: a list of Bin object
    """
    if data is None or len(data) <= 0:
        raise ValueError('Empty input array!')

    if n_bins <= 0:
        raise ValueError('Less than one bin makes no sense.')

    insufficient_distinct = False
    n_unique_values = len(np.unique([value for value in data if not is_nan(value)]))
    if n_unique_values < n_bins:
        insufficient_distinct = True
        warnings.warn("Insufficient unique values for requested number of bins. " +
                      "Number of bins will be reset to number of unique values.")
        n_bins = n_unique_values

    # cast into a numpy array to infer the dtype
    data_as_array = np.array(data)
    is_numeric = np.issubdtype(data_as_array.dtype, np.number)

    if is_numeric:
        bins = _create_numerical_bins(data_as_array, n_bins)
    else:
        bins = _create_categorical_bins(data_as_array, n_bins)

    if (not insufficient_distinct) and (len(bins) < n_bins):
        warnings.warn('Created less bins than requested.')

    return bins


#------- private methods for numerical binnings-------#

def _create_numerical_bins(data_as_array, n_bins):
    return _create_next_numerical_bin(data_as_array, n_bins, [])


def _create_next_numerical_bin(x, n_bins, result):
    """
    Create bins for numerical data
    :param x: array of data
    :param n_bins: number of bins
    :return: a list of bins object
    """
    # no more data
    if len(x) == 0:
        return result

    # if data has nan
    if any(np.isnan(x)):
        cur_bin = Bin("numerical", np.nan, np.nan, True, True)
        result.append(cur_bin)
        return _create_next_numerical_bin(x[~np.isnan(x)], n_bins-1, result)

    # the last bin is a closed-closed interval
    if n_bins == 1:
        cur_bin = Bin("numerical", min(x), max(x), True, True)
        result.append(cur_bin)
        return result

    lower, upper, lower_closed, upper_closed = _first_interval(x, n_bins)
    cur_bin = Bin("numerical", lower, upper, lower_closed, upper_closed)
    result.append(cur_bin)

    next_data = x[x > upper] if upper_closed else x[x >= upper]
    return _create_next_numerical_bin(next_data, n_bins-1, result)


def _first_interval(x, n_bins):
    """
    Gets the first interval based on the percentiles, 
    either a closed interval containing the same value multiple times
    or a closed-open interval with a different lower and upper bound.
    """
    # calculate the percentiles
    percentiles = np.linspace(0., 100., n_bins + 1)
    bounds = np.percentile(x, q=percentiles, interpolation='higher')
    lower = bounds[0]
    upper = bounds[1]

    if lower == upper:
        return lower, upper, True, True
    else:
        return lower, upper, True, False


#------- private methods for categorical binnings-------#

def _create_categorical_bins(data_as_array, n_bins):
    """ 
    Performs greedy (non-optimal) binning
    :param data_as_array: array of data to bin according to their frequencies
    :return: a list of Bin object
    """
    # count items
    weights = {}
    for item in data_as_array:
        if not item in weights:
            weights[item] = 1
        else:
            weights[item] += 1

    # we need items sorted in decreasing order
    pairs = sorted([(weight, [item]) for (item, weight) in weights.items()], reverse=True)
    bins = pairs[:min(n_bins, len(pairs))]

    # too little data, just return what we have so far
    if len(pairs) <= n_bins:
        return toBinObject(bins)

    heapify(bins)

    # go through pairs, from heaviest to lightest
    for (pair_weight, pair_labels) in pairs[n_bins:]:
        # take the lightest bin
        bin_weight, bin_labels = heappop(bins)
        # add the heaviest item to it
        heappush(bins, (bin_weight + pair_weight, bin_labels + pair_labels))

    return toBinObject(bins)


def toBinObject(bins):
    result = []
    for bin in bins:
        categories = bin[1]
        result.append(Bin("categorical", categories))
    return result
